import logging
from typing import Set

from fastapi import APIRouter
from fastapi import Query
from fastapi import Request

from catweazle.authorize import Authorize
from catweazle.errors import SessionCredentialError

from catweazle.crud.instances import CrudInstances

from catweazle.model.v2.instances import filter_list
from catweazle.model.v2.instances import filter_literal
from catweazle.model.v1.instances import ModelV1InstanceGet


class ControllerApiV1Instances:

    def __init__(
        self,
        log: logging.Logger,
        authorize: Authorize,
        crud_instances: CrudInstances,
    ):
        self._authorize = authorize
        self._crud_instances = crud_instances
        self._log = log
        self._router = APIRouter(
            prefix="/instances",
            tags=["instances"],
        )
        self.router.add_api_route(
            "/{instance_id}",
            self.get,
            response_model=ModelV1InstanceGet,
            response_model_exclude_unset=True,
            methods=["GET"],
        )

    @property
    def authorize(self):
        return self._authorize

    @property
    def crud_instances(self):
        return self._crud_instances

    @property
    def log(self):
        return self._log

    @property
    def router(self):
        return self._router

    async def get(
        self,
        instance_id: str,
        request: Request,
        fields: Set[filter_literal] = Query(default=filter_list),
    ):
        try:
            await self.authorize.require_user(request=request)
            fields.discard("ipa_otp")
        except SessionCredentialError as err:
            result = await self.crud_instances.get(
                _id=instance_id, fields=["id", "ip_address"]
            )
            if result.ip_address != request.client.host:
                raise err

        result = await self.crud_instances.get(_id=instance_id, fields=list(fields))
        return ModelV1InstanceGet(data=result)



import logging

import httpx
from fastapi import APIRouter

from catweazle.controller.oauth.authenticate import ControllerOauthAuthenticate
from catweazle.crud.oauth import CrudOAuth
from catweazle.crud.users import CrudUsers


class ControllerOauth:
    def __init__(
        self,
        log: logging.Logger,
        curd_oauth: dict[str, CrudOAuth],
        crud_users: CrudUsers,
        http: httpx.AsyncClient,
    ):
        self._log = log
        self._router = APIRouter()

        self.router.include_router(
            ControllerOauthAuthenticate(
                log=log,
                crud_oauth=curd_oauth,
                crud_users=crud_users,
                http=http,
            ).router,
            responses={404: {"description": "Not found"}},
        )

    @property
    def router(self):
        return self._router



import pymongo


class FilterMixIn(object):
    @staticmethod
    def _filter_boolean(query, field, selector):
        if selector is None:
            return
        if selector in [True, "true", "True", "1"]:
            selector = True
        else:
            selector = False
        query[field] = selector

    @staticmethod
    def _filter_list(query, field, selector, nin=False):
        if selector is None:
            return
        if type(selector) is not list:
            selector = list(set(selector.split(",")))
        if nin:
            query[field] = {"$nin": selector}
        else:
            query[field] = {"$in": selector}

    @staticmethod
    def _filter_re(query, field, selector, list_filter=None):
        if selector and list_filter is not None:
            query[field] = {"$regex": selector, "$in": list_filter}
        elif selector:
            query[field] = {"$regex": selector}
        elif list_filter is not None:
            query[field] = {"$in": list_filter}

    @staticmethod
    def _filter_literal(query, field, selector, list_filter=None):
        if selector and list_filter:
            query[field] = {"$eq": selector, "$in": list_filter}
        elif selector:
            query[field] = selector
        elif list_filter:
            query[field] = {"$in": list_filter}


class Format:
    @staticmethod
    def _format(item):
        item.pop("_id", None)
        return item

    @staticmethod
    def _format_multi(item, count=None):
        return {"result": item, "meta": {"result_size": count}}


class PaginationSkipMixIn:
    @staticmethod
    def _pagination_skip(page, limit):
        return page * limit


class ProjectionMixIn:
    @staticmethod
    def _projection(fields):
        if not fields:
            return None
        result = {}
        for field in fields:
            result[field] = 1
        return result


class SortMixIn:
    @staticmethod
    def _sort(sort, sort_order):
        if sort_order == "ascending":
            return [(sort, pymongo.ASCENDING)]
        else:
            return [(sort, pymongo.DESCENDING)]



import logging
from typing import List

import httpx
from fastapi import APIRouter

from catweazle.authorize import Authorize

from catweazle.controller.api import ControllerApi
from catweazle.controller.oauth import ControllerOauth

from catweazle.crud.credentials import CrudCredentials
from catweazle.crud.ldap import CrudLdap
from catweazle.crud.foreman import CrudForeman
from catweazle.crud.instances import CrudInstances
from catweazle.crud.oauth import CrudOAuth
from catweazle.crud.permissions import CrudPermissions
from catweazle.crud.users import CrudUsers


class Controller:
    def __init__(
        self,
        log: logging.Logger,
        authorize: Authorize,
        crud_ldap: CrudLdap,
        crud_foreman_backends: List[CrudForeman],
        crud_instances: CrudInstances,
        crud_oauth: dict[str, CrudOAuth],
        crud_permissions: CrudPermissions,
        crud_users: CrudUsers,
        crud_users_credentials: CrudCredentials,
        http: httpx.AsyncClient,
    ):
        self._log = log
        self._router = APIRouter()

        self.router.include_router(
            ControllerApi(
                log=log,
                authorize=authorize,
                crud_ldap=crud_ldap,
                crud_instances=crud_instances,
                crud_permissions=crud_permissions,
                crud_users=crud_users,
                crud_users_credentials=crud_users_credentials,
                crud_foreman_backends=crud_foreman_backends,
                http=http,
            ).router,
            prefix="/api",
            responses={404: {"description": "Not found"}},
        )

        self.router.include_router(
            ControllerOauth(
                log=log,
                curd_oauth=crud_oauth,
                crud_users=crud_users,
                http=http,
            ).router,
            prefix="/oauth",
            responses={404: {"description": "Not found"}},
        )

    @property
    def router(self):
        return self._router

    @property
    def log(self):
        return self._log



import re
from typing import Literal
from typing import Set

from pydantic import BaseModel
from pydantic import constr
from pydantic import Field
from typing_extensions import Annotated


sort_order_literal = Literal[
    "ascending",
    "descending",
]


class ModelV2MetaMulti(BaseModel):
    result_size: Annotated[int, Field(gt=-1)]


class ModelV2DataDelete(BaseModel):
    pass



import logging
from typing import List
from typing import Set

from fastapi import APIRouter
from fastapi import Query
from fastapi import Request

from catweazle.authorize import Authorize

from catweazle.errors import SessionCredentialError

from catweazle.crud.instances import CrudInstances
from catweazle.crud.foreman import CrudForeman
from catweazle.errors import BackendError

from catweazle.model.v2.common import ModelV2DataDelete
from catweazle.model.v2.common import sort_order_literal
from catweazle.model.v2.instances import filter_list
from catweazle.model.v2.instances import filter_literal
from catweazle.model.v2.instances import sort_literal
from catweazle.model.v2.instances import ModelV2InstanceGet
from catweazle.model.v2.instances import ModelV2InstanceGetMulti
from catweazle.model.v2.instances import ModelV2instancePost
from catweazle.model.v2.instances import ModelV2instancePut


class ControllerApiV2Instances:

    def __init__(
        self,
        log: logging.Logger,
        authorize: Authorize,
        crud_instances: CrudInstances,
        crud_foreman_backends: List[CrudForeman],
    ):
        self._authorize = authorize
        self._crud_instances = crud_instances
        self._crud_foreman_backends = crud_foreman_backends
        self._log = log
        self._router = APIRouter(
            prefix="/instances",
            tags=["instances"],
        )

        self.router.add_api_route(
            "",
            self.search,
            response_model=ModelV2InstanceGetMulti,
            response_model_exclude_unset=True,
            methods=["GET"],
        )
        self.router.add_api_route(
            "/{instance_id}",
            self.create,
            response_model=ModelV2InstanceGet,
            response_model_exclude_unset=True,
            methods=["POST"],
            status_code=201,
        )
        self.router.add_api_route(
            "/{instance_id}",
            self.delete,
            response_model=ModelV2DataDelete,
            response_model_exclude_unset=True,
            methods=["DELETE"],
        )
        self.router.add_api_route(
            "/{instance_id}",
            self.get,
            response_model=ModelV2InstanceGet,
            response_model_exclude_unset=True,
            methods=["GET"],
        )
        self.router.add_api_route(
            "/{instance_id}",
            self.update,
            response_model=ModelV2InstanceGet,
            response_model_exclude_unset=True,
            methods=["PUT"],
        )

    @property
    def authorize(self):
        return self._authorize

    @property
    def crud_instances(self):
        return self._crud_instances

    @property
    def crud_foreman_backends(self):
        return self._crud_foreman_backends

    @property
    def log(self):
        return self._log

    @property
    def router(self):
        return self._router

    async def create(
        self,
        data: ModelV2instancePost,
        instance_id: str,
        request: Request,
        fields: Set[filter_literal] = Query(default=filter_list),
    ):
        await self.authorize.require_permission(
            request=request, permission="INSTANCE:POST"
        )

        instance = await self.crud_instances.create(
            _id=instance_id, payload=data, fields=list(fields)
        )
        for foreman in self.crud_foreman_backends:
            try:
                await foreman.create_dns(
                    fqdn=instance.fqdn,
                    ip_address=instance.ip_address,
                )
                await foreman.create_realm(
                    fqdn=instance.fqdn,
                )
            except BackendError as err:
                self.log.error(
                    f"Failed to create DNS or realm for instance {instance_id} in foreman backend {foreman.name}"
                )
                await self.delete(
                    instance_id=instance_id,
                    request=request,
                )
                raise err
        return instance

    async def delete(self, request: Request, instance_id: str):
        await self.authorize.require_permission(
            request=request, permission="INSTANCE:DELETE"
        )
        instance = await self.crud_instances.get(
            _id=instance_id, fields=["fqdn", "ip_address"]
        )
        for foreman in self.crud_foreman_backends:
            try:
                await foreman.delete_dns(
                    fqdn=instance.fqdn,
                    ip_address=instance.ip_address,
                )
            except BackendError:
                pass
            try:
                await foreman.delete_realm(
                    fqdn=instance.fqdn,
                )
            except BackendError:
                pass
        return await self.crud_instances.delete(_id=instance_id)

    async def get(
        self,
        instance_id: str,
        request: Request,
        fields: Set[filter_literal] = Query(default=filter_list),
    ):
        try:
            await self.authorize.require_user(request=request)
            fields.discard("ipa_otp")
        except SessionCredentialError as err:
            result = await self.crud_instances.get(
                _id=instance_id, fields=["id", "ip_address"]
            )
            if result.ip_address != request.client.host:
                raise err

        return await self.crud_instances.get(_id=instance_id, fields=list(fields))

    async def search(
        self,
        request: Request,
        instance_id: str = Query(
            description="filter: regular_expressions", default=None
        ),
        dns_indicator: str = Query(
            description="filter: regular_expressions", default=None
        ),
        ip_address: str = Query(
            description="filter: regular_expressions", default=None
        ),
        fqdn: str = Query(description="filter: regular_expressions", default=None),
        fields: Set[filter_literal] = Query(default=filter_list),
        sort: sort_literal = Query(default="id"),
        sort_order: sort_order_literal = Query(default="ascending"),
        page: int = Query(default=0, ge=0, description="pagination index"),
        limit: int = Query(
            default=10,
            ge=10,
            le=1000,
            description="pagination limit, min value 10, max value 1000",
        ),
    ):
        await self.authorize.require_user(request=request)
        fields.discard("ipa_otp")
        return await self.crud_instances.search(
            _id=instance_id,
            dns_indicator=dns_indicator,
            ip_address=ip_address,
            fqdn=fqdn,
            fields=list(fields),
            sort=sort,
            sort_order=sort_order,
            page=page,
            limit=limit,
        )

    async def update(
        self,
        data: ModelV2instancePut,
        instance_id: str,
        request: Request,
        fields: Set[filter_literal] = Query(default=filter_list),
    ):
        await self.authorize.require_permission(
            request=request, permission="INSTANCE:POST"
        )
        return await self.crud_instances.update(
            _id=instance_id, payload=data, fields=list(fields)
        )



import logging
from typing import Set

from fastapi import APIRouter
from fastapi import Query
from fastapi import Request

from catweazle.authorize import Authorize

from catweazle.crud.permissions import CrudPermissions
from catweazle.crud.ldap import CrudLdap

from catweazle.model.v2.common import ModelV2DataDelete
from catweazle.model.v2.common import sort_order_literal
from catweazle.model.v2.permissions import filter_list
from catweazle.model.v2.permissions import filter_literal
from catweazle.model.v2.permissions import sort_literal
from catweazle.model.v2.permissions import ModelV2PermissionGet
from catweazle.model.v2.permissions import ModelV2PermissionGetMulti
from catweazle.model.v2.permissions import ModelV2PermissionPost
from catweazle.model.v2.permissions import ModelV2PermissionPut


class ControllerApiV2Permissions:
    def __init__(
        self,
        log: logging.Logger,
        authorize: Authorize,
        crud_permissions: CrudPermissions,
        crud_ldap: CrudLdap,
    ):
        self._authorize = authorize
        self._crud_permissions = crud_permissions
        self._crud_ldap = crud_ldap
        self._log = log
        self._router = APIRouter(
            prefix="/permissions",
            tags=["permissions"],
        )

        self.router.add_api_route(
            "",
            self.search,
            response_model=ModelV2PermissionGetMulti,
            response_model_exclude_unset=True,
            methods=["GET"],
        )
        self.router.add_api_route(
            "/{permission_id}",
            self.create,
            response_model=ModelV2PermissionGet,
            response_model_exclude_unset=True,
            methods=["POST"],
            status_code=201,
        )
        self.router.add_api_route(
            "/{permission_id}",
            self.delete,
            response_model=ModelV2DataDelete,
            response_model_exclude_unset=True,
            methods=["DELETE"],
        )
        self.router.add_api_route(
            "/{permission_id}",
            self.get,
            response_model=ModelV2PermissionGet,
            response_model_exclude_unset=True,
            methods=["GET"],
        )
        self.router.add_api_route(
            "/{permission_id}",
            self.update,
            response_model=ModelV2PermissionGet,
            response_model_exclude_unset=True,
            methods=["PUT"],
        )

    @property
    def authorize(self):
        return self._authorize

    @property
    def crud_permissions(self):
        return self._crud_permissions

    @property
    def crud_ldap(self):
        return self._crud_ldap

    @property
    def log(self):
        return self._log

    @property
    def router(self):
        return self._router

    async def create(
        self,
        request: Request,
        data: ModelV2PermissionPost,
        permission_id: str,
        fields: Set[filter_literal] = Query(default=filter_list),
    ):
        await self.authorize.require_admin(request=request)
        if data.ldap_group:
            data.users = await self.crud_ldap.get_logins_from_group(
                group=data.ldap_group
            )
        return await self.crud_permissions.create(
            _id=permission_id,
            payload=data,
            fields=list(fields),
        )

    async def delete(
        self,
        request: Request,
        permission_id: str,
    ):
        await self.authorize.require_admin(request=request)
        return await self.crud_permissions.delete(
            _id=permission_id,
        )

    async def get(
        self,
        permission_id: str,
        request: Request,
        fields: Set[filter_literal] = Query(default=filter_list),
    ):
        await self.authorize.require_admin(request=request)
        return await self.crud_permissions.get(_id=permission_id, fields=list(fields))

    async def search(
        self,
        request: Request,
        permission_id: str = Query(
            description="filter: regular_expressions", default=None
        ),
        ldap_group: str = Query(
            description="filter: regular_expressions", default=None
        ),
        users: str = Query(description="filter: regular_expressions", default=None),
        fields: Set[filter_literal] = Query(default=filter_list),
        sort: sort_literal = Query(default="id"),
        sort_order: sort_order_literal = Query(default="ascending"),
        page: int = Query(default=0, ge=0, description="pagination index"),
        limit: int = Query(
            default=10,
            ge=10,
            le=1000,
            description="pagination limit, min value 10, max value 1000",
        ),
    ):
        await self.authorize.require_admin(request=request)
        return await self.crud_permissions.search(
            _id=permission_id,
            ldap_group=ldap_group,
            users=users,
            fields=list(fields),
            sort=sort,
            sort_order=sort_order,
            page=page,
            limit=limit,
        )

    async def update(
        self,
        data: ModelV2PermissionPut,
        permission_id: str,
        request: Request,
        fields: Set[filter_literal] = Query(default=filter_list),
    ):
        await self.authorize.require_admin(request=request)
        current_group = await self.crud_permissions.get(
            _id=permission_id,
            fields=["ldap_group", "users"],
        )
        if data.ldap_group:
            data.users = await self.crud_ldap.get_logins_from_group(
                group=data.ldap_group
            )
        elif current_group.ldap_group:
            data.users = await self.crud_ldap.get_logins_from_group(
                group=current_group.ldap_group
            )
        return await self.crud_permissions.update(
            _id=permission_id,
            payload=data,
            fields=list(fields),
        )



from typing import get_args as typing_get_args
from typing import List
from typing import Literal
from typing import Optional
from typing_extensions import Annotated
from pydantic import BaseModel
from pydantic import StrictStr
from pydantic import StringConstraints

from catweazle.model.v2.common import ModelV2MetaMulti

filter_literal = Literal[
    "id",
    "ldap_group",
    "permissions",
    "users",
]

filter_list = set(typing_get_args(filter_literal))

sort_literal = Literal["id"]

Permissions = Annotated[
    str,
    StringConstraints(pattern="^(INSTANCE:(DELETE|POST))$"),
]


class ModelV2PermissionGet(BaseModel):
    id: Optional[StrictStr] = None
    ldap_group: Optional[StrictStr] = ""
    permissions: Optional[List[Permissions]] = None
    users: Optional[List[StrictStr]] = None


class ModelV2PermissionGetMulti(BaseModel):
    result: List[ModelV2PermissionGet]
    meta: ModelV2MetaMulti


class ModelV2PermissionPost(BaseModel):
    ldap_group: Optional[StrictStr] = ""
    permissions: Optional[List[Permissions]] = None
    users: Optional[List[StrictStr]] = []


class ModelV2PermissionPut(BaseModel):
    ldap_group: Optional[StrictStr] = None
    permissions: Optional[List[Permissions]] = None
    users: Optional[List[StrictStr]] = None



import logging
import json
import os
import sys
import time

import boto3
import botocore.exceptions
import httpx

__version__ = "2.0.0"


def lambda_handler(event, context):
    cat_weazle_lambda = CatWeazleLambda(context=context, event=event)
    cat_weazle_lambda.run()


class CatWeazleLambda(object):
    def __init__(self, event, context):
        self._boto = None
        self.context = context
        self.event = event
        self._logger = logging.getLogger()
        self.log.setLevel(os.environ.get("CatWeazleLogLevel", logging.INFO))
        self._api_version = None
        self.cw_endpoint = os.environ.get("CatWeazleEndPoint")
        self.cw_indicator_tag = os.environ.get("CatWeazleIndicatorTag")
        self.cw_indicator_tmpl = os.environ.get("CatWeazleIndicatorTemplate", None)
        self.cw_role_arn = os.environ.get("CatWeazleRoleARN")
        self.cw_secret = os.environ.get("CatWeazleSecret")
        self.cw_secret_id = os.environ.get("CatWeazleSecretID")
        self.cw_name_tag = os.environ.get("CatWeazleNameTag", "Name")
        self.cw_name_tag_if_empty = os.environ.get("CatWeazleNameTagIfEmpty", True)
        self.cw_name_target_tag = os.environ.get("CatWeazleNameTargetTag", None)
        self.cw_role_session_name = os.environ.get(
            "CatWeazleRoleSessionName", "catweazle_session"
        )
        self.cw_post_create_lambda = os.environ.get("CatWeazlePostCreateLambda", None)
        self._instance = None
        self._validate_configuration()

    def _validate_configuration(self):
        required_vars = {
            "CatWeazleEndPoint": self.cw_endpoint,
            "CatWeazleSecret": self.cw_secret,
            "CatWeazleSecretID": self.cw_secret_id,
            "CatWeazleRoleARN": self.cw_role_arn,
            "CatWeazleIndicatorTag": self.cw_indicator_tag,
        }

        missing = [name for name, value in required_vars.items() if not value]
        if missing:
            self.log.fatal(
                f"Missing required environment variables: {', '.join(missing)}"
            )
            sys.exit(0)

    @property
    def boto(self):
        if not self._boto:
            access_role_arn = self.cw_role_arn.format(self.account_id)
            self.log.info(self._fmt_log_msg(f"Assuming role {access_role_arn}"))
            self._boto = self._get_session(access_role_arn)
        return self._boto

    @property
    def api_version(self):
        if not self._api_version:
            try:
                data = self.catweazle_api(
                    method="GET",
                    endpoint="version",
                )
                self._api_version = data["version"]
            except httpx.HTTPError as err:
                self.log.warning(
                    self._fmt_log_msg(f"Could not get API version: {err}, assuming v1")
                )
                self._api_version = "v1"
        return self._api_version

    @property
    def ec2_id(self):
        return self.event["detail"]["instance-id"]

    @property
    def account_id(self):
        return self.event["account"]

    @property
    def instance(self):
        if not self._instance:
            ec2 = self.boto.resource("ec2")
            self._instance = ec2.Instance(self.ec2_id)
        return self._instance

    @property
    def log(self):
        return self._logger

    def _get_role_credentials(self, role):
        session = boto3.Session()
        sts = session.client("sts")
        role = sts.assume_role(
            RoleArn=role,
            RoleSessionName=self.cw_role_session_name,
            DurationSeconds=900,
        )
        access_key = role["Credentials"]["AccessKeyId"]
        secret_key = role["Credentials"]["SecretAccessKey"]
        session_token = role["Credentials"]["SessionToken"]
        return access_key, secret_key, session_token

    def _get_session(self, role):
        cred = self._get_role_credentials(role)
        session = boto3.Session(
            aws_access_key_id=cred[0],
            aws_secret_access_key=cred[1],
            aws_session_token=cred[2],
        )
        return session

    def catweazle_api(self, method, endpoint, body=None, api_version=None) -> dict:
        url = f"{self.cw_endpoint}/api/{endpoint}"
        headers = {
            "X-ID": self.cw_secret_id,
            "X-SECRET-ID": self.cw_secret_id,
            "X-SECRET": self.cw_secret,
        }
        if api_version == "v1" and body:
            body = {"data": body}
        try:
            self.log.info(self._fmt_log_msg(f"request: {method} {url} {body}"))
            with httpx.Client(timeout=30) as client:
                resp = client.request(method, url, headers=headers, json=body)
                resp.raise_for_status()
                if api_version == "v1" and method != "DELETE":
                    return resp.json()["data"]
                return resp.json()
        except httpx.HTTPError as err:
            self.log.error(self._fmt_log_msg(f"request error: {err}"))
            raise err

    def post_create_lambda(self):
        self.log.info(self._fmt_log_msg("calling post create lambda functions"))
        if not self.cw_post_create_lambda:
            self.log.info(
                self._fmt_log_msg("calling post create lambda functions, done")
            )
            return
        self.log.info(self._fmt_log_msg("getting session in this account"))
        session = boto3.Session()
        aws_lambda = session.client("lambda")
        self.log.info(self._fmt_log_msg("getting session in this account, done"))
        for function in self.cw_post_create_lambda.split(","):
            try:
                self.log.info(
                    self._fmt_log_msg(f"calling post create lambda function {function}")
                )
                response = aws_lambda.invoke(
                    FunctionName=function,
                    InvocationType="Event",
                    Payload=json.dumps(self.event),
                )
                self.log.info(
                    self._fmt_log_msg(
                        f"calling post create lambda function {function}, done, response: {response}"
                    )
                )
            except Exception as err:
                self.log.info(
                    self._fmt_log_msg(
                        f"calling post create lambda function {function}, failed, error: {err}"
                    )
                )
        self.log.info(self._fmt_log_msg("calling post create lambda functions, done"))

    def _add_body_meta(self, body):
        if self.api_version == "v1":
            return
        body["meta"] = dict()
        body["meta"]["account_id"] = self.account_id
        body["meta"]["instance_id"] = self.ec2_id
        body["meta"]["instance_state"] = self.event["detail"]["state"]

    def instance_create(self):
        self.log.info(self._fmt_log_msg("create instance"))
        body = dict()
        body["ip_address"] = self.instance.private_ip_address
        body["dns_indicator"] = self.get_dns_indicator()
        self._add_body_meta(body)
        try:
            path = f"{self.api_version}/instances/{self.ec2_id}"
            resp = self.catweazle_api(
                method="POST", endpoint=path, body=body, api_version=self.api_version
            )
        except httpx.HTTPError as err:
            self.log.error(self._fmt_log_msg(f"create instance failed {err}"))
            sys.exit(1)
        fqdn = resp["fqdn"]
        fqdn_tag_name = self.get_fqdn_tag_name()
        self.set_ec2_tag(fqdn=fqdn, fqdn_tag_name=fqdn_tag_name)
        if not self.get_tag("Name") and self.cw_name_tag_if_empty:
            self.log.info(
                self._fmt_log_msg("setting Name tag, since it is currently empty.")
            )
            self.set_ec2_tag(fqdn=fqdn, fqdn_tag_name="Name")
        self.post_create_lambda()
        self.log.info(self._fmt_log_msg("create instance, done"))

    def instance_delete(self):
        self.log.info(self._fmt_log_msg("deleting instance"))
        try:
            path = f"{self.api_version}/instances/{self.ec2_id}"
            self.catweazle_api(
                method="DELETE", endpoint=path, api_version=self.api_version
            )
        except httpx.HTTPError as err:
            self.log.error(self._fmt_log_msg(f"could not delete instance {err}"))
            sys.exit(0)
        self.log.info(self._fmt_log_msg("deleting instance, done"))

    def instance_update(self):
        if self.api_version == "v1":
            return
        self.log.info(self._fmt_log_msg("update instance"))
        body = dict()
        self._add_body_meta(body)
        try:
            path = f"{self.api_version}/instances/{self.ec2_id}"
            self.catweazle_api(
                method="PUT", endpoint=path, body=body, api_version=self.api_version
            )
        except httpx.HTTPError as err:
            self.log.error(self._fmt_log_msg(f"could not update instance {err}"))
            sys.exit(0)
        self.log.info(self._fmt_log_msg("update instance, done"))

    def set_ec2_tag(self, fqdn, fqdn_tag_name):
        try:
            self.log.info(
                self._fmt_log_msg(f"setting {fqdn_tag_name} tag of instance to {fqdn}")
            )
            self.instance.create_tags(Tags=[{"Key": fqdn_tag_name, "Value": fqdn}])
            self.log.info(
                self._fmt_log_msg(
                    f"setting {fqdn_tag_name} tag of instance to {fqdn}, done"
                )
            )
        except botocore.exceptions.ClientError as err:
            self.log.error(
                self._fmt_log_msg(f"setting {fqdn_tag_name} tag failed {err}")
            )

    def get_dns_indicator(self):
        self.log.info(self._fmt_log_msg("fetching instance dns indicator"))
        dns_indicator = self.get_tag(self.cw_indicator_tag, retry=10)
        if not dns_indicator:
            self.log.error(
                self._fmt_log_msg("instance is missing the name indicator tag, exit")
            )
            sys.exit(0)
        if dns_indicator.startswith("INSTANCEID"):
            dns_indicator = dns_indicator.replace("INSTANCEID", self.ec2_id)
        if self.cw_indicator_tmpl:
            dns_indicator = self.cw_indicator_tmpl.format(dns_indicator)
        self.log.info(self._fmt_log_msg(f"using indicator {dns_indicator}"))
        self.log.info(self._fmt_log_msg("fetching instance dns indicator, done"))
        return dns_indicator

    def get_fqdn_tag_name(self):
        fqdn_tag_name = self.cw_name_tag
        self.log.info(self._fmt_log_msg(f"setting Name tag to default {fqdn_tag_name}"))
        if self.cw_name_target_tag:
            name_target_tag = self.get_tag(self.cw_name_target_tag)
            if name_target_tag:
                self.log.info(
                    self._fmt_log_msg(f"overriding Name tag to {name_target_tag}")
                )
                return name_target_tag
        return fqdn_tag_name

    def get_tag(self, tag_name, retry=None):
        if not retry:
            return self._get_tag(tag_name)
        while retry > 0:
            result = self._get_tag(tag_name)
            if result:
                return result
            self.instance.reload()
            retry -= 1
            time.sleep(2)
        return None

    def _get_tag(self, tag_name):
        if not self.instance.tags:
            return None
        for tag in self.instance.tags:
            if tag["Key"] == tag_name:
                return tag["Value"]
        return None

    def _fmt_log_msg(self, msg):
        return f"{self.account_id}/{self.ec2_id}: {msg}"

    def run(self):
        self.log.info(
            f"script version: {__version__}, start working on {self.event['id']}, event payload: {self.event}"
        )

        if self.event["source"] != "aws.ec2":
            self.log.fatal(f"got event from unexpected event source: {self.event}")
            sys.exit(0)

        state = self.event["detail"]["state"]

        self.log.info(self._fmt_log_msg(f"got aws.ec2 {state} event"))

        if state == "pending":
            self.instance_create()
        elif state == "shutting-down":
            self.instance_update()
            self.instance_delete()
        elif state == "terminated":
            self.instance_update()
            self.instance_delete()
        else:
            self.instance_update()

        self.log.info(self._fmt_log_msg(f"finished working on {self.event['id']}"))



import logging
from typing import List

import httpx
from fastapi import APIRouter

from catweazle.authorize import Authorize

from catweazle.controller.api.v1 import ControllerApiV1
from catweazle.controller.api.v2 import ControllerApiV2

from catweazle.crud.credentials import CrudCredentials
from catweazle.crud.ldap import CrudLdap
from catweazle.crud.foreman import CrudForeman
from catweazle.crud.instances import CrudInstances
from catweazle.crud.permissions import CrudPermissions
from catweazle.crud.users import CrudUsers

from catweazle.model import ModelApiVersions


class ControllerApi:
    def __init__(
        self,
        log: logging.Logger,
        authorize: Authorize,
        crud_ldap: CrudLdap,
        crud_foreman_backends: List[CrudForeman],
        crud_instances: CrudInstances,
        crud_permissions: CrudPermissions,
        crud_users: CrudUsers,
        crud_users_credentials: CrudCredentials,
        http: httpx.AsyncClient,
    ):
        self._router = APIRouter()
        self._log = log

        self.router.include_router(
            ControllerApiV2(
                log=log,
                authorize=authorize,
                crud_ldap=crud_ldap,
                crud_foreman_backends=crud_foreman_backends,
                crud_instances=crud_instances,
                crud_permissions=crud_permissions,
                crud_users=crud_users,
                crud_users_credentials=crud_users_credentials,
                http=http,
            ).router,
            prefix="/v2",
            responses={404: {"description": "Not found"}},
        )

        self.router.include_router(
            ControllerApiV1(
                log=log,
                authorize=authorize,
                crud_instances=crud_instances,
            ).router,
            prefix="/v1",
            responses={404: {"description": "Not found"}},
        )

        self.router.add_api_route(
            "/version",
            self.get_api_versions,
            response_model=ModelApiVersions,
            methods=["GET"],
            tags=["api_version"],
        )

    @property
    def router(self):
        return self._router

    @property
    def log(self):
        return self._log

    @staticmethod
    async def get_api_versions() -> ModelApiVersions:
        return ModelApiVersions(version="v2")



from typing import get_args as typing_get_args
from typing import Optional
from typing import List
from typing import Literal

from pydantic import BaseModel

from catweazle.model.v2.common import ModelV2MetaMulti

filter_literal = Literal[
    "id",
    "created",
    "description",
]

filter_list = set(typing_get_args(filter_literal))

sort_literal = Literal[
    "id",
    "created",
]


class ModelV2CredentialGet(BaseModel):
    id: Optional[str] = None
    created: Optional[str] = None
    description: Optional[str] = None


class ModelV2CredentialGetMulti(BaseModel):
    result: List[ModelV2CredentialGet]
    meta: ModelV2MetaMulti


class ModelV2CredentialPost(BaseModel):
    description: str


class ModelV2CredentialPostResult(ModelV2CredentialGet):
    secret: str


class ModelV2CredentialPut(BaseModel):
    description: str



from contextlib import asynccontextmanager
import logging
import random
import string
import sys
import time
from typing import List

from authlib.integrations.starlette_client import OAuth
import bonsai.asyncio
import httpx
from fastapi import FastAPI
from motor.motor_asyncio import AsyncIOMotorClient
from motor.motor_asyncio import AsyncIOMotorDatabase
from starlette.middleware.sessions import SessionMiddleware
import uvicorn

import catweazle.controller
import catweazle.controller.oauth

from catweazle.authorize import Authorize

from catweazle.config import Config
from catweazle.config import ConfigLdap as SettingsLdap
from catweazle.config import ConfigOAuth as SettingsOAuth

from catweazle.crud.credentials import CrudCredentials
from catweazle.crud.ldap import CrudLdap
from catweazle.crud.foreman import CrudForeman
from catweazle.crud.instances import CrudInstances
from catweazle.crud.oauth import CrudOAuthGitHub
from catweazle.crud.permissions import CrudPermissions
from catweazle.crud.users import CrudUsers

from catweazle.model.v2.users import ModelV2UserPost

from catweazle.errors import ResourceNotFound


settings = Config()


@asynccontextmanager
async def lifespan(app: FastAPI):
    log = setup_logging(
        settings.app.loglevel,
    )

    http = httpx.AsyncClient()

    ldap_pool = await setup_ldap(
        log=log,
        settings_ldap=settings.ldap,
    )

    log.info("adding routes")
    mongo_db = setup_mongodb(
        log=log,
        database=settings.mongodb.database,
        url=settings.mongodb.url,
    )

    crud_foreman_backends = setup_foreman_backend(log=log)

    crud_oauth = setup_oauth_providers(
        log=log,
        http=http,
        oauth_settings=settings.oauth,
    )

    crud_ldap = CrudLdap(
        log=log,
        ldap_base_dn=settings.ldap.basedn,
        ldap_bind_dn=settings.ldap.binddn,
        ldap_pool=ldap_pool,
        ldap_url=settings.ldap.url,
        ldap_user_pattern=settings.ldap.userpattern,
    )

    crud_instances = CrudInstances(
        log=log,
        coll=mongo_db["instances"],
        domain_suffix=settings.app.domainsuffix,
    )
    await crud_instances.index_create()

    crud_permissions = CrudPermissions(
        log=log,
        coll=mongo_db["permissions"],
    )
    await crud_permissions.index_create()

    crud_users = CrudUsers(
        log=log,
        coll=mongo_db["users"],
        crud_ldap=crud_ldap,
    )
    await crud_users.index_create()

    crud_users_credentials = CrudCredentials(
        log=log,
        coll=mongo_db["users_credentials"],
    )
    await crud_users_credentials.index_create()

    authorize = Authorize(
        log=log,
        crud_permissions=crud_permissions,
        crud_users=crud_users,
        crud_users_credentials=crud_users_credentials,
    )

    controller = catweazle.controller.Controller(
        log=log,
        authorize=authorize,
        crud_ldap=crud_ldap,
        crud_foreman_backends=crud_foreman_backends,
        crud_instances=crud_instances,
        crud_permissions=crud_permissions,
        crud_users=crud_users,
        crud_users_credentials=crud_users_credentials,
        crud_oauth=crud_oauth,
        http=http,
    )
    app.include_router(controller.router)

    log.info("adding routes, done")
    await setup_admin_user(log=log, crud_users=crud_users)
    yield


async def setup_admin_user(log: logging.Logger, crud_users: CrudUsers):
    try:
        await crud_users.get(_id="admin", fields=["_id"])
    except ResourceNotFound:
        password = "".join(
            random.choice(string.ascii_letters + string.digits) for _ in range(20)
        )
        log.info(f"creating admin user with password {password}")
        await crud_users.create(
            _id="admin",
            payload=ModelV2UserPost(
                admin=True,
                email="admin@example.com",
                name="admin",
                password=password,
            ),
            fields=["_id"],
        )
        log.info("creating admin user, done")


async def setup_ldap(log: logging.Logger, settings_ldap: SettingsLdap):
    if not settings_ldap.url:
        log.info("ldap not configured")
        return
    log.info(f"setting up ldap with {settings_ldap.url} as a backend")
    if not settings_ldap.binddn:
        log.fatal("ldap binddn not configured")
        sys.exit(1)
    if not settings_ldap.password:
        log.fatal("ldap password not configured")
        sys.exit(1)
    client = bonsai.LDAPClient(settings_ldap.url)
    client.set_credentials("SIMPLE", settings_ldap.binddn, settings_ldap.password)
    pool = bonsai.asyncio.AIOConnectionPool(client=client, maxconn=30)
    await pool.open()
    return pool


def setup_foreman_backend(log: logging.Logger) -> List[CrudForeman]:
    if not settings.foreman:
        log.info("no foreman backend configured")
    backends = list()
    realm = False
    if not settings.foreman:
        return backends
    for name, config in settings.foreman.items():
        log.info(f"setting up foreman backend with name {name}")
        if realm and config.realmenable:
            log.fatal("only one realm provider is allowed")
            sys.exit(1)
        if config.realmenable:
            log.info(f"realm provider enabled for {name}")
            realm = True
        backends.append(
            CrudForeman(
                log=log,
                name=name,
                url=config.url,
                ssl_key=config.sslkey,
                ssl_ca=config.sslca,
                ssl_crt=config.sslcrt,
                dns_arpa_enable=config.dnsarpaenable,
                dns_arpa_zones=config.dnsarpazones,
                dns_forward_enable=config.dnsforwardenable,
                realm_enable=config.realmenable,
                realm_name=config.realmname,
            )
        )
    return backends


def setup_logging(log_level):
    log = logging.getLogger("uvicorn")
    log.info(f"setting loglevel to: {log_level}")
    log.setLevel(log_level)
    return log


def setup_mongodb(log: logging.Logger, database: str, url: str) -> AsyncIOMotorDatabase:
    log.info("setting up mongodb client")
    pool = AsyncIOMotorClient(url)
    db = pool.get_database(database)
    log.info("setting up mongodb client, done")
    return db


def setup_oauth_providers(
    log: logging.Logger,
    http: httpx.AsyncClient,
    oauth_settings: dict["str", SettingsOAuth],
):
    oauth = OAuth()
    providers = {}
    if not oauth_settings:
        log.info("oauth not configured")
        return providers
    for provider, config in oauth_settings.items():
        if config.type == "github":
            log.info(f"oauth setting up github provider with name {provider}")
            providers[provider] = CrudOAuthGitHub(
                log=log,
                http=http,
                backend_override=config.override,
                name=provider,
                oauth=oauth,
                scope=config.scope,
                client_id=config.client.id,
                client_secret=config.client.secret,
                authorize_url=config.url.authorize,
                access_token_url=config.url.accesstoken,
                userinfo_url=config.url.userinfo,
            )
    return providers


app = FastAPI(title="catweazle", version="0.0.0", lifespan=lifespan)
app.add_middleware(SessionMiddleware, secret_key=settings.app.secretkey, max_age=3600)


@app.middleware("http")
async def add_process_time_header(request, call_next):
    start_time = time.time()
    response = await call_next(request)
    process_time = time.time() - start_time
    response.headers["X-Process-Time"] = str(process_time)
    return response


def main():
    uvicorn.run(
        app,
        host=settings.app.host,
        port=settings.app.port,
        proxy_headers=settings.app.proxyheaders,
        forwarded_allow_ips=settings.app.forwardedallowips,
    )


if __name__ == "__main__":
    main()



import logging
import typing

from bson.objectid import ObjectId
from motor.motor_asyncio import AsyncIOMotorCollection
import pymongo
import pymongo.errors

from catweazle.crud.mixins import FilterMixIn
from catweazle.crud.mixins import Format
from catweazle.crud.mixins import PaginationSkipMixIn
from catweazle.crud.mixins import ProjectionMixIn
from catweazle.crud.mixins import SortMixIn

from catweazle.errors import DuplicateResource
from catweazle.errors import ResourceNotFound
from catweazle.errors import BackendError


class Crud:
    def __init__(self, log: logging.Logger):
        self._log = log

    @property
    def log(self):
        return self._log


class CrudMongo(
    Crud, FilterMixIn, Format, PaginationSkipMixIn, ProjectionMixIn, SortMixIn
):
    def __init__(self, log: logging.Logger, coll: AsyncIOMotorCollection):
        super().__init__(log)
        self._resource_type = coll.name
        self._coll = coll

    @property
    def coll(self):
        return self._coll

    @property
    def resource_type(self):
        return self._resource_type

    async def _create(
        self,
        payload: dict,
        fields: list = None,
    ) -> dict:
        try:
            _id = await self._coll.insert_one(payload)
            return await self._get_by_obj_id(_id=_id.inserted_id, fields=fields)
        except pymongo.errors.DuplicateKeyError:
            raise DuplicateResource
        except pymongo.errors.ConnectionFailure as err:
            self.log.error(f"backend error: {err}")
            raise BackendError()

    async def _delete(self, query: dict) -> dict:
        try:
            result = await self._coll.delete_one(filter=query)
        except pymongo.errors.ConnectionFailure as err:
            self.log.error(f"backend error: {err}")
            raise BackendError()
        if result.deleted_count == 0:
            raise ResourceNotFound
        return {}

    async def _get(self, query: dict, fields: list) -> dict:
        try:
            result = await self._coll.find_one(
                filter=query, projection=self._projection(fields)
            )
        except pymongo.errors.ConnectionFailure as err:
            self.log.error(f"backend error: {err}")
            raise BackendError
        if result is None:
            if "$and" in query:
                query = query["$and"][0]
            raise ResourceNotFound(
                details=f"Resource {self.resource_type} {query} not found"
            )
        return self._format(result)

    async def _get_by_obj_id(self, _id, fields: list) -> dict:
        query = {"_id": _id}
        return await self._get(query=query, fields=fields)

    async def _resource_exists(self, query: dict) -> ObjectId:
        result = await self._get(query=query, fields=["id"])
        return result["id"]

    async def _search(
        self,
        query: dict,
        fields: typing.Optional[list] = None,
        sort: typing.Optional[str] = None,
        sort_order: typing.Optional[str] = None,
        page: typing.Optional[int] = None,
        limit: typing.Optional[int] = None,
    ) -> dict:
        try:
            count = await self._coll.count_documents(
                filter=query,
            )
            cursor = self._coll.find(filter=query, projection=self._projection(fields))
            if sort and sort_order:
                cursor.sort(self._sort(sort=sort, sort_order=sort_order))
            if page and limit:
                cursor.skip(self._pagination_skip(page, limit))
                cursor.limit(limit)
            return self._format_multi(list(await cursor.to_list(limit)), count=count)
        except pymongo.errors.ConnectionFailure as err:
            self.log.error(f"backend error: {err}")
            raise BackendError

    async def _update(
        self, query: dict, payload: dict, fields: list, upsert=False
    ) -> dict:
        update = {"$set": {}}
        for k, v in payload.items():
            if v is None:
                continue
            update["$set"][k] = v
        try:
            result = await self._coll.find_one_and_update(
                filter=query,
                update=update,
                projection=self._projection(fields=fields),
                return_document=pymongo.ReturnDocument.AFTER,
                upsert=upsert,
            )
        except pymongo.errors.ConnectionFailure as err:
            self.log.error(f"backend error: {err}")
            raise BackendError
        if result is None:
            raise ResourceNotFound
        return self._format(result)



import logging
from typing import List
import ssl
import sys

from catweazle.errors import BackendError

import httpx
import ipaddress
from pydantic.networks import IPv4Network


class CrudForeman:
    def __init__(
        self,
        log: logging.Logger,
        name: str,
        url: str,
        ssl_key: str,
        ssl_ca: str,
        ssl_crt: str,
        dns_arpa_enable: bool,
        dns_arpa_zones: List[IPv4Network],
        dns_forward_enable: bool,
        realm_enable: bool = False,
        realm_name: str = None,
    ):
        self._log = log
        self._dns_arpa_enable = dns_arpa_enable
        self._dns_arpa_zones = set()
        for subnet in dns_arpa_zones:
            self._dns_arpa_zones.add(ipaddress.ip_network(subnet))
        self._dns_forward_enable = dns_forward_enable
        self._name = name
        self._realm_enable = realm_enable
        self._realm_name = realm_name
        self._ssl_ca = ssl_ca
        self._ssl_crt = ssl_crt
        self._ssl_key = ssl_key
        self._url = url
        if self.ssl_key and self.ssl_crt and self.ssl_ca:
            context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
            try:
                context.load_cert_chain(certfile=self.ssl_crt, keyfile=self.ssl_key)
                context.load_verify_locations(cafile=self.ssl_ca)
            except OSError as err:
                self.log.error(
                    f"{self.name}:foreman: could not create ssl context: {err}"
                )
                sys.exit(1)
            self._http = httpx.AsyncClient(verify=context, timeout=30)
        else:
            self._http = httpx.AsyncClient(timeout=30)

    @property
    def log(self):
        return self._log

    @property
    def dns_arpa_enable(self):
        return self._dns_arpa_enable

    @property
    def dns_arpa_zones(self):
        return self._dns_arpa_zones

    @property
    def dns_forward_enable(self):
        return self._dns_forward_enable

    @property
    def name(self):
        return self._name

    @property
    def realm_enable(self):
        return self._realm_enable

    @property
    def realm_name(self):
        return self._realm_name

    @property
    def ssl_ca(self):
        return self._ssl_ca

    @property
    def ssl_crt(self):
        return self._ssl_crt

    @property
    def ssl_key(self):
        return self._ssl_key

    @property
    def url(self):
        return self._url

    @property
    def http(self) -> httpx.AsyncClient:
        return self._http

    async def request_delete(self, url):
        url = f"{self.url}{url}"
        resp = await self.http.delete(url)
        self.log.info(f"{self.name}:foreman: request_delete: {url}")
        if resp.status_code != 200:
            self.log.error(resp.text)
            raise BackendError()

    async def request_post(self, url, data):
        url = f"{self.url}{url}"
        resp = await self.http.post(url, data=data)
        self.log.info(f"{self.name}:foreman: request_post: {url} {data}")
        if resp.status_code != 200:
            self.log.error(resp.text)
            raise BackendError()
        return resp.json()

    def arpa_responsible(self, ip_address):
        for zone in self.dns_arpa_zones:
            if ip_address in zone:
                return True
        return False

    async def create_dns(self, fqdn, ip_address):
        await self.create_arpa_dns(fqdn=fqdn, ip_address=ip_address)
        await self.create_forward_dns(fqdn=fqdn, ip_address=ip_address)

    async def create_arpa_dns(self, fqdn, ip_address):
        self.log.info(
            f"{self.name}:foreman: creating DNS PTR Record for {fqdn} with ip {ip_address}"
        )
        if not self.dns_arpa_enable:
            self.log.info(f"{self.name}:foreman: reverse DNS is disabled")
            return
        ip_addr = ipaddress.IPv4Address(ip_address)
        if self.arpa_responsible(ip_address=ip_addr):
            body_ptr = {
                "fqdn": fqdn,
                "value": ip_addr.reverse_pointer,
                "type": "PTR",
            }
            await self.request_post("/dns/", body_ptr)
        self.log.info(
            f"{self.name}:foreman: creating DNS PTR Record for {fqdn} with ip {ip_address}, done"
        )

    async def create_forward_dns(self, fqdn, ip_address):
        self.log.info(
            f"{self.name}:foreman: creating DNS A Record for {fqdn} with ip {ip_address}"
        )
        if not self.dns_forward_enable:
            self.log.info(f"{self.name}:foreman: forward DNS is disabled")
            return
        body_a = {
            "fqdn": fqdn,
            "value": ip_address,
            "type": "A",
        }
        self.log.info(
            f"{self.name}:foreman: creating DNS A Record for {fqdn} with ip {ip_address}, done"
        )

        await self.request_post("/dns/", body_a)

    async def delete_dns(self, fqdn, ip_address):
        await self.delete_arpa_dns(fqdn=fqdn, ip_address=ip_address)
        await self.delete_forward_dns(fqdn=fqdn, ip_address=ip_address)

    async def delete_arpa_dns(self, ip_address, fqdn):
        self.log.info(
            f"{self.name}:foreman: deleting DNS PTR Record for {fqdn} with ip {ip_address}"
        )
        if not self.dns_arpa_enable:
            self.log.info(f"{self.name}:foreman: reverse DNS is disabled")
            return
        ip_addr = ipaddress.IPv4Address(ip_address)
        if self.arpa_responsible(ip_address=ip_addr):
            await self.request_delete(f"/dns/{ip_addr.reverse_pointer}/PTR")
        self.log.info(
            f"{self.name}:foreman: deleting DNS PTR Record for {fqdn} with ip {ip_address}, done"
        )

    async def delete_forward_dns(self, ip_address, fqdn):
        self.log.info(
            f"{self.name}:foreman: deleting DNS A Record for {fqdn} with ip {ip_address}"
        )
        if not self.dns_forward_enable:
            self.log.info(f"{self.name}:foreman: forward DNS is disabled")
            return
        await self.request_delete(f"/dns/{fqdn}/A")
        self.log.info(
            f"{self.name}:foreman: deleting DNS A Record for {fqdn} with ip {ip_address}, done"
        )

    async def create_realm(self, fqdn):
        if not self.realm_enable:
            self.log.info(f"{self.name}:foreman: realm is disabled")
            return None
        self.log.info(f"{self.name}:foreman: creating realm entry for {fqdn}")
        body_a = {
            "hostname": fqdn,
        }
        data = await self.request_post(f"/realm/{self.realm_name}", body_a)
        self.log.info(f"{self.name}:foreman: creating realm entry for {fqdn}, done")
        return data["randompassword"]

    async def delete_realm(self, fqdn):
        if not self.realm_enable:
            self.log.info(f"{self.name}:foreman: realm is disabled")
            return
        self.log.info(f"{self.name}:foreman: deleting realm entry for {fqdn}")
        self.log.info(f"{self.name}:foreman: deleting realm entry for {fqdn}, done")
        await self.request_delete(f"/realm/{self.realm_name}/{fqdn}")



from typing import get_args as typing_get_args
from typing import Literal
from pydantic import BaseModel

from catweazle.model.v2.instances import ModelV2InstanceGet
from catweazle.config import Config

config = Config()

filter_literal = Literal[
    "id",
    "dns_indicator",
    "fqdn",
    "ip_address",
    "ipa_otp",
]

filter_list = set(typing_get_args(filter_literal))

sort_literal = Literal[
    "id",
    "dns_indicator",
    "fqdn",
    "ip_address",
]


class ModelV1InstanceGet(BaseModel):
    data: ModelV2InstanceGet



import logging

from fastapi import APIRouter

from catweazle.authorize import Authorize

from catweazle.controller.api.v1.instances import ControllerApiV1Instances

from catweazle.crud.instances import CrudInstances


class ControllerApiV1:
    def __init__(
        self,
        log: logging.Logger,
        authorize: Authorize,
        crud_instances: CrudInstances,
    ):
        self._router = APIRouter()
        self._log = log

        self.router.include_router(
            ControllerApiV1Instances(
                log=log,
                authorize=authorize,
                crud_instances=crud_instances,
            ).router,
            responses={404: {"description": "Not found"}},
        )

    @property
    def router(self):
        return self._router



from fastapi import HTTPException


class AuthenticationError(HTTPException):
    def __init__(self, msg="Invalid Username and Password"):
        super(AuthenticationError, self).__init__(status_code=401, detail=msg)


class DuplicateResource(HTTPException):
    def __init__(self):
        super(DuplicateResource, self).__init__(
            status_code=400, detail="Duplicate Resource"
        )


class ResourceNotFound(HTTPException):
    def __init__(self, details=None):
        if not details:
            details = "Resource not found"
        super(ResourceNotFound, self).__init__(status_code=404, detail=details)


class BackendError(HTTPException):
    def __init__(self):
        super(BackendError, self).__init__(
            status_code=500,
            detail="Internal Server Error: please contact administrator",
        )


class LdapResourceNotFound(HTTPException):
    def __init__(self):
        super(LdapResourceNotFound, self).__init__(
            status_code=400, detail="Ldap Resource not Found"
        )


class LdapInvalidDN(HTTPException):
    def __init__(self):
        super(LdapInvalidDN, self).__init__(status_code=400, detail="invalid ldap dn")


class LdapNoBackend(HTTPException):
    def __init__(self):
        super(LdapNoBackend, self).__init__(
            status_code=400,
            detail="No Ldap Backend configured, ldap operations not supported",
        )


class AdminError(HTTPException):
    def __init__(self):
        super(AdminError, self).__init__(
            status_code=403, detail="Admin Permissions required"
        )


class CredentialError(HTTPException):
    def __init__(self):
        super(CredentialError, self).__init__(
            status_code=403, detail="Invalid or no credentials"
        )


class SessionCredentialError(HTTPException):
    def __init__(self):
        super(SessionCredentialError, self).__init__(
            status_code=401, detail="No Session or API Credentials present"
        )


class PermError(HTTPException):
    def __init__(self, permission):
        super(PermError, self).__init__(
            status_code=403,
            detail=f"Permissions error, you are not granted {permission} on this resource",
        )


class HostNumRangeExceeded(HTTPException):
    def __init__(self, max_range):
        super(HostNumRangeExceeded, self).__init__(
            status_code=400,
            detail=f"Host number range exceeded for indicator {max_range}, "
            "please use a different indicator or remove existing hosts",
        )



import logging
import typing
from bson.objectid import ObjectId
from motor.motor_asyncio import AsyncIOMotorCollection
import pymongo
import pymongo.errors

from catweazle.crud.common import CrudMongo

from catweazle.model.v2.common import ModelV2DataDelete
from catweazle.model.v2.common import sort_order_literal
from catweazle.model.v2.permissions import ModelV2PermissionGet
from catweazle.model.v2.permissions import ModelV2PermissionGetMulti
from catweazle.model.v2.permissions import ModelV2PermissionPost
from catweazle.model.v2.permissions import ModelV2PermissionPut


class CrudPermissions(CrudMongo):
    def __init__(self, log: logging.Logger, coll: AsyncIOMotorCollection):
        super(CrudPermissions, self).__init__(log=log, coll=coll)

    async def index_create(self) -> None:
        self.log.info(f"creating {self.resource_type} indices")
        await self.coll.create_index(
            [
                ("id", pymongo.ASCENDING),
            ],
            unique=True,
        )
        await self.coll.create_index(
            [
                ("ldap_group", pymongo.ASCENDING),
            ]
        )
        await self.coll.create_index(
            [
                ("users", pymongo.ASCENDING),
            ]
        )
        self.log.info(f"creating {self.resource_type} indices, done")

    async def create(
        self,
        _id: str,
        payload: ModelV2PermissionPost,
        fields: list,
    ) -> ModelV2PermissionGet:
        data = payload.model_dump()
        data["id"] = _id
        result = await self._create(payload=data, fields=fields)
        return ModelV2PermissionGet(**result)

    async def delete(
        self,
        _id: str,
    ) -> ModelV2DataDelete:
        query = {"id": _id}
        await self._delete(query=query)
        return ModelV2DataDelete()

    async def delete_user_from_permissions(self, user_id):
        query = {}
        update = {"$pull": {"users": user_id}}
        await self._coll.update_many(
            filter=query,
            update=update,
        )

    async def get(
        self,
        _id: str,
        fields: list,
    ) -> ModelV2PermissionGet:
        query = {"id": _id}
        result = await self._get(query=query, fields=fields)
        return ModelV2PermissionGet(**result)

    async def resource_exists(
        self,
        _id: str,
    ) -> ObjectId:
        query = {"id": _id}
        return await self._resource_exists(query=query)

    async def search(
        self,
        _id: typing.Optional[str] = None,
        ldap_group: typing.Optional[str] = None,
        permissions: typing.Optional[str] = None,
        users: typing.Optional[str] = None,
        fields: typing.Optional[list] = None,
        sort: typing.Optional[str] = None,
        sort_order: typing.Optional[sort_order_literal] = None,
        page: typing.Optional[int] = None,
        limit: typing.Optional[int] = None,
    ) -> ModelV2PermissionGetMulti:
        query = {}
        self._filter_re(query, "id", _id)
        self._filter_re(query, "ldap_group", ldap_group)
        self._filter_re(query, "permissions", permissions)
        self._filter_re(query, "users", users)

        result = await self._search(
            query=query,
            fields=fields,
            sort=sort,
            sort_order=sort_order,
            page=page,
            limit=limit,
        )
        return ModelV2PermissionGetMulti(**result)

    async def update(
        self,
        _id: str,
        payload: ModelV2PermissionPut,
        fields: list,
    ) -> ModelV2PermissionGet:
        query = {"id": _id}
        data = payload.model_dump()

        result = await self._update(query=query, fields=fields, payload=data)
        return ModelV2PermissionGet(**result)



import logging


from fastapi import APIRouter
from fastapi import Request

import httpx

from catweazle.authorize import Authorize
from catweazle.crud.users import CrudUsers

from catweazle.errors import AuthenticationError

from catweazle.model.v2.common import ModelV2DataDelete
from catweazle.model.v2.authenticate import ModelV2AuthenticateGetUser
from catweazle.model.v2.authenticate import ModelV2AuthenticatePost


class ControllerApiV2Authenticate:
    def __init__(
        self,
        log: logging.Logger,
        authorize: Authorize,
        crud_users: CrudUsers,
        http: httpx.AsyncClient,
    ):
        self._authorize = authorize
        self._crud_users = crud_users
        self._http = http
        self._log = log
        self._router = APIRouter(
            prefix="/authenticate",
            tags=["authenticate"],
        )

        self.router.add_api_route(
            "", self.get, response_model=ModelV2AuthenticateGetUser, methods=["GET"]
        )
        self.router.add_api_route(
            "",
            self.create,
            response_model=ModelV2AuthenticateGetUser,
            methods=["POST"],
            status_code=201,
        )
        self.router.add_api_route(
            "", self.delete, response_model=ModelV2DataDelete, methods=["DELETE"]
        )

    @property
    def authorize(self):
        return self._authorize

    @property
    def crud_users(self):
        return self._crud_users

    @property
    def http(self):
        return self._http

    @property
    def log(self):
        return self._log

    @property
    def router(self):
        return self._router

    async def get(self, request: Request):
        user = await self.authorize.get_user(request=request)
        if not user:
            raise AuthenticationError
        return {"user": user.id}

    async def create(
        self,
        data: ModelV2AuthenticatePost,
        request: Request,
    ):
        user = await self.crud_users.check_credentials(credentials=data)
        request.session["username"] = user
        return {"user": user}

    @staticmethod
    async def delete(request: Request):
        request.session.clear()
        return {}



from typing import List
from typing import Optional
from pydantic import BaseModel
from pydantic import StrictStr

from catweazle.model.v2.common import ModelV2MetaMulti


class ModelOauthProviderGet(BaseModel):
    id: Optional[StrictStr] = None


class ModelOauthProviderGetMulti(BaseModel):
    result: List[ModelOauthProviderGet]
    meta: ModelV2MetaMulti



import typing

from pydantic import BaseModel
from pydantic import StrictStr
from pydantic import field_validator
from pydantic.networks import IPv4Network
from pydantic_settings import BaseSettings
from pydantic_settings import SettingsConfigDict

log_levels = typing.Literal[
    "CRITICAL", "FATAL", "ERROR", "WARN", "WARNING", "INFO", "DEBUG"
]


class ConfigApp(BaseModel):
    loglevel: log_levels = "INFO"
    host: str = "127.0.0.1"
    port: int = 8000
    secretkey: str = "secret"
    indicatorregex: str = "^.*NUM.*$"
    domainsuffix: str = ".example.com"
    proxyheaders: typing.Optional[bool] = False
    forwardedallowips: typing.Optional[str] = "127.0.0.0/8"


class ConfigLdap(BaseModel):
    url: typing.Optional[str] = None
    basedn: typing.Optional[str] = None
    binddn: typing.Optional[str] = None
    password: typing.Optional[str] = None
    userpattern: typing.Optional[str] = None


class ConfigMongodb(BaseModel):
    url: str = "mongodb://localhost:27017"
    database: str = "catweazle"


class ConfigOAuthClient(BaseModel):
    id: str
    secret: str


class ConfigOAuthUrl(BaseModel):
    authorize: str
    accesstoken: str
    userinfo: typing.Optional["str"] = None


class ConfigOAuth(BaseModel):
    override: bool = False
    scope: str
    type: str
    client: ConfigOAuthClient
    url: ConfigOAuthUrl


class ConfigForeman(BaseModel):
    url: StrictStr
    sslca: typing.Optional[StrictStr] = None
    sslcrt: typing.Optional[StrictStr] = None
    sslkey: typing.Optional[StrictStr] = None
    dnsforwardenable: bool = False
    dnsarpaenable: bool = False
    dnsarpazones: typing.Optional[typing.List[IPv4Network]] = []
    realmenable: bool = False
    realmname: typing.Optional[StrictStr] = None

    @field_validator("dnsarpazones", mode="before")
    def validate_dns_indicator(v):
        networks = list()
        for network in v.split():
            try:
                networks.append(IPv4Network(network.strip()))
            except ValueError as e:
                raise ValueError(f"Invalid IPv4 network: {network}") from e
        return networks


class Config(BaseSettings):
    app: ConfigApp = ConfigApp()
    ldap: ConfigLdap = ConfigLdap()
    mongodb: ConfigMongodb = ConfigMongodb()
    foreman: typing.Optional[dict[str, ConfigForeman]] = None
    oauth: typing.Optional[dict[str, ConfigOAuth]] = None
    model_config = SettingsConfigDict(env_file=".env", env_nested_delimiter="_")



import logging
from typing import Set

from fastapi import APIRouter
from fastapi import Query
from fastapi import Request

from catweazle.authorize import Authorize

from catweazle.crud.permissions import CrudPermissions
from catweazle.crud.users import CrudUsers
from catweazle.crud.credentials import CrudCredentials

from catweazle.model.v2.common import ModelV2DataDelete
from catweazle.model.v2.common import sort_order_literal
from catweazle.model.v2.users import filter_list
from catweazle.model.v2.users import filter_literal
from catweazle.model.v2.users import sort_literal
from catweazle.model.v2.users import ModelV2UserGet
from catweazle.model.v2.users import ModelV2UserGetMulti
from catweazle.model.v2.users import ModelV2UserPost
from catweazle.model.v2.users import ModelV2UserPut


class ControllerApiV2Users:
    def __init__(
        self,
        log: logging.Logger,
        authorize: Authorize,
        crud_permissions: CrudPermissions,
        crud_users: CrudUsers,
        crud_users_credentials: CrudCredentials,
    ):
        self._authorize = authorize
        self._crud_permissions = crud_permissions
        self._crud_users = crud_users
        self._crud_users_credentials = crud_users_credentials
        self._log = log
        self._router = APIRouter(
            prefix="/users",
            tags=["users"],
        )

        self.router.add_api_route(
            "",
            self.search,
            response_model=ModelV2UserGetMulti,
            response_model_exclude_unset=True,
            methods=["GET"],
        )
        self.router.add_api_route(
            "/{user_id}",
            self.create,
            response_model=ModelV2UserGet,
            response_model_exclude_unset=True,
            methods=["POST"],
            status_code=201,
        )
        self.router.add_api_route(
            "/{user_id}",
            self.delete,
            response_model=ModelV2DataDelete,
            response_model_exclude_unset=True,
            methods=["DELETE"],
        )
        self.router.add_api_route(
            "/{user_id}",
            self.get,
            response_model=ModelV2UserGet,
            response_model_exclude_unset=True,
            methods=["GET"],
        )
        self.router.add_api_route(
            "/{user_id}",
            self.update,
            response_model=ModelV2UserGet,
            response_model_exclude_unset=True,
            methods=["PUT"],
        )

    @property
    def authorize(self):
        return self._authorize

    @property
    def crud_permissions(self):
        return self._crud_permissions

    @property
    def crud_users(self):
        return self._crud_users

    @property
    def curd_users_credentials(self):
        return self._crud_users_credentials

    @property
    def log(self):
        return self._log

    @property
    def router(self):
        return self._router

    async def create(
        self,
        request: Request,
        data: ModelV2UserPost,
        user_id: str,
        fields: Set[filter_literal] = Query(default=filter_list),
    ):
        await self.authorize.require_admin(request=request)
        return await self.crud_users.create(
            _id=user_id, payload=data, fields=list(fields)
        )

    async def delete(self, request: Request, user_id: str):
        await self.authorize.require_admin(request=request)
        await self.curd_users_credentials.delete_all_from_owner(owner=user_id)
        await self.crud_permissions.delete_user_from_permissions(user_id=user_id)
        return await self.crud_users.delete(_id=user_id)

    async def get(
        self,
        user_id: str,
        request: Request,
        fields: Set[filter_literal] = Query(default=filter_list),
    ):
        if user_id == "_self":
            user_id = await self.authorize.get_user(request=request)
            user_id = user_id.id
        else:
            await self.authorize.require_admin(request=request)
        return await self.crud_users.get(_id=user_id, fields=list(fields))

    async def search(
        self,
        request: Request,
        user_id: str = Query(description="filter: regular_expressions", default=None),
        fields: Set[filter_literal] = Query(default=filter_list),
        sort: sort_literal = Query(default="id"),
        sort_order: sort_order_literal = Query(default="ascending"),
        page: int = Query(default=0, ge=0, description="pagination index"),
        limit: int = Query(
            default=10,
            ge=10,
            le=1000,
            description="pagination limit, min value 10, max value 1000",
        ),
    ):
        await self.authorize.require_admin(request=request)
        return await self.crud_users.search(
            _id=user_id,
            fields=list(fields),
            sort=sort,
            sort_order=sort_order,
            page=page,
            limit=limit,
        )

    async def update(
        self,
        data: ModelV2UserPut,
        user_id: str,
        request: Request,
        fields: Set[filter_literal] = Query(default=filter_list),
    ):
        if user_id == "_self":
            user_id = await self.authorize.get_user(request=request)
            user_id = user_id.id
            data.admin = None
        else:
            await self.authorize.require_admin(request=request)
        return await self.crud_users.update(
            _id=user_id, payload=data, fields=list(fields)
        )



from datetime import datetime
from datetime import UTC
import logging
import random
import string
import typing
import uuid

from fastapi import Request
from motor.motor_asyncio import AsyncIOMotorCollection
from passlib.hash import pbkdf2_sha512
import pymongo

from catweazle.crud.common import CrudMongo

from catweazle.errors import CredentialError
from catweazle.errors import ResourceNotFound

from catweazle.model.v2.common import ModelV2DataDelete
from catweazle.model.v2.common import sort_order_literal
from catweazle.model.v2.credentials import ModelV2CredentialGet
from catweazle.model.v2.credentials import ModelV2CredentialGetMulti
from catweazle.model.v2.credentials import ModelV2CredentialPost
from catweazle.model.v2.credentials import ModelV2CredentialPostResult
from catweazle.model.v2.credentials import ModelV2CredentialPut


class CrudCredentials(CrudMongo):
    def __init__(self, log: logging.Logger, coll: AsyncIOMotorCollection):
        super(CrudCredentials, self).__init__(log=log, coll=coll)

    @staticmethod
    def _create_secret(token) -> str:
        return pbkdf2_sha512.encrypt(str(token), rounds=10, salt_size=32)

    async def index_create(self) -> None:
        self.log.info(f"creating {self.resource_type} indices")
        await self.coll.create_index(
            [("id", pymongo.ASCENDING), ("owner", pymongo.ASCENDING)], unique=True
        )
        self.log.info(f"creating {self.resource_type} indices, done")

    async def check_credential(self, request: Request):
        x_secret = request.headers.get("x-secret")
        x_secret_id = request.headers.get("x-secret-id")
        if not x_secret_id:
            x_secret_id = request.headers.get("x-id")

        query = {"id": x_secret_id}

        result = await self._get(query=query, fields=["secret", "owner"])

        if not pbkdf2_sha512.verify(x_secret, result["secret"]):
            raise CredentialError

        return result["owner"]

    async def create(
        self,
        owner: str,
        payload: ModelV2CredentialPost,
    ) -> ModelV2CredentialPostResult:
        data = payload.model_dump()
        _id = uuid.uuid4()
        secret = "".join(
            random.SystemRandom().choice(string.ascii_letters + string.digits + "_-.")
            for _ in range(128)
        )
        created = datetime.now(UTC)
        data["id"] = str(_id)
        data["secret"] = self._create_secret(str(secret))
        data["created"] = created
        data["owner"] = owner
        await self._create(payload=data, fields=["id"])
        result = {
            "id": str(_id),
            "created": str(created),
            "description": payload.description,
            "secret": str(secret),
        }
        return ModelV2CredentialPostResult(**result)

    async def delete(self, _id: str, owner: str) -> ModelV2DataDelete:
        query = {"id": _id, "owner": owner}
        await self._delete(query=query)
        return ModelV2DataDelete()

    async def delete_all_from_owner(self, owner: str) -> ModelV2DataDelete:
        query = {"owner": owner}
        try:
            await self._delete(query=query)
        except ResourceNotFound:
            pass
        return ModelV2DataDelete()

    async def get(self, _id: str, owner: str, fields: list) -> ModelV2CredentialGet:
        query = {"id": str(_id), "owner": owner}
        result = await self._get(query=query, fields=fields)
        if "created" in result:
            result["created"] = str(result["created"])
        self.log.info(result)
        return ModelV2CredentialGet(**result)

    async def search(
        self,
        owner: typing.Optional[str] = None,
        fields: typing.Optional[list] = None,
        sort: typing.Optional[str] = None,
        sort_order: typing.Optional[sort_order_literal] = None,
        page: typing.Optional[int] = None,
        limit: typing.Optional[int] = None,
    ) -> ModelV2CredentialGetMulti:
        query = {"owner": owner}

        result = await self._search(
            query=query,
            fields=fields,
            sort=sort,
            sort_order=sort_order,
            page=page,
            limit=limit,
        )
        for item in result["result"]:
            if "created" in item:
                item["created"] = str(item["created"])
        self.log.info(result)
        return ModelV2CredentialGetMulti(**result)

    async def update(
        self, _id: str, owner: str, payload: ModelV2CredentialPut, fields: list
    ) -> ModelV2CredentialGet:
        query = {"id": _id, "owner": owner}
        data = payload.model_dump()
        result = await self._update(query=query, fields=fields, payload=data)
        if "created" in result:
            result["created"] = str(result["created"])
        return ModelV2CredentialGet(**result)



import logging
import typing

from bson.objectid import ObjectId
from motor.motor_asyncio import AsyncIOMotorCollection
from passlib.hash import pbkdf2_sha512
import pymongo
import pymongo.errors

from catweazle.crud.common import CrudMongo
from catweazle.crud.ldap import CrudLdap

from catweazle.errors import AuthenticationError
from catweazle.errors import BackendError

from catweazle.model.v2.common import ModelV2DataDelete
from catweazle.model.v2.common import sort_order_literal
from catweazle.model.v2.authenticate import ModelV2AuthenticatePost
from catweazle.model.v2.users import ModelV2UserGet
from catweazle.model.v2.users import ModelV2UserGetMulti
from catweazle.model.v2.users import ModelV2UserPost
from catweazle.model.v2.users import ModelV2UserPut


class CrudUsers(CrudMongo):
    def __init__(
        self,
        log: logging.Logger,
        coll: AsyncIOMotorCollection,
        crud_ldap: CrudLdap,
    ):
        super(CrudUsers, self).__init__(log=log, coll=coll)
        self._crud_ldap = crud_ldap

    async def index_create(self) -> None:
        self.log.info(f"creating {self.resource_type} indices")
        await self.coll.create_index([("id", pymongo.ASCENDING)], unique=True)
        self.log.info(f"creating {self.resource_type} indices, done")

    @property
    def crud_ldap(self):
        return self._crud_ldap

    @staticmethod
    def _password(password) -> str:
        return pbkdf2_sha512.encrypt(password, rounds=100000, salt_size=32)

    async def check_credentials(self, credentials: ModelV2AuthenticatePost) -> str:
        user = credentials.user
        password = credentials.password
        try:
            result = await self._coll.find_one(
                filter={"id": user},
                projection={"password": 1, "backend": 1},
            )
            if not result:
                await self.check_credentials_ldap_and_create_user(
                    credentials=credentials
                )
            elif result["backend"] == "internal":
                if not pbkdf2_sha512.verify(password, result["password"]):
                    raise AuthenticationError
            elif result["backend"] == "ldap":
                try:
                    await self.crud_ldap.check_user_credentials(
                        user=user, password=password
                    )
                except AuthenticationError:
                    raise AuthenticationError
            else:
                self.log.error(
                    f"auth backend mismatch, expected ldap or internal, got: {result['backend']}"
                )
                raise AuthenticationError(
                    msg="backend mismatch, please contact the administrator"
                )
            return user
        except pymongo.errors.ConnectionFailure as err:
            self.log.error(f"backend error: {err}")
            raise BackendError()

    async def check_credentials_ldap_and_create_user(
        self, credentials: ModelV2AuthenticatePost
    ):
        ldap_user = await self.crud_ldap.check_user_credentials(
            user=credentials.user,
            password=credentials.password,
        )
        result = await self.create_external(
            _id=credentials.user,
            payload=ModelV2UserPut(
                name=f"{ldap_user['givenName'][0]} {ldap_user['sn'][0]}",
                email=ldap_user["mail"][0],
                admin=False,
            ),
            backend="ldap",
            fields=["_id"],
        )
        return result

    async def create(
        self,
        _id: str,
        payload: ModelV2UserPost,
        fields: list,
    ) -> ModelV2UserGet:
        data = payload.model_dump()
        data["id"] = _id
        data["password"] = self._password(payload.password)
        data["backend"] = "internal"
        result = await self._create(payload=data, fields=fields)
        return ModelV2UserGet(**result)

    async def create_external(
        self,
        _id: str,
        payload: ModelV2UserPut,
        fields: list,
        backend: str,
    ) -> ModelV2UserGet:
        data = payload.model_dump()
        data["id"] = _id
        data["backend"] = backend
        result = await self._create(payload=data, fields=fields)
        return ModelV2UserGet(**result)

    async def delete(
        self,
        _id: str,
    ) -> ModelV2DataDelete:
        query = {"id": _id}
        await self._delete(query=query)
        return ModelV2DataDelete()

    async def get(
        self,
        _id: str,
        fields: list,
    ) -> ModelV2UserGet:
        query = {"id": _id}
        result = await self._get(query=query, fields=fields)
        return ModelV2UserGet(**result)

    async def resource_exists(
        self,
        _id: str,
    ) -> ObjectId:
        query = {"id": _id}
        return await self._resource_exists(query=query)

    async def search(
        self,
        _id: typing.Optional[str] = None,
        fields: typing.Optional[list] = None,
        sort: typing.Optional[str] = None,
        sort_order: typing.Optional[sort_order_literal] = None,
        page: typing.Optional[int] = None,
        limit: typing.Optional[int] = None,
    ) -> ModelV2UserGetMulti:
        query = {}
        self._filter_re(query, "id", _id)

        result = await self._search(
            query=query,
            fields=fields,
            sort=sort,
            sort_order=sort_order,
            page=page,
            limit=limit,
        )
        return ModelV2UserGetMulti(**result)

    async def update(
        self,
        _id: str,
        payload: ModelV2UserPut,
        fields: list,
    ) -> ModelV2UserGet:
        query = {"id": _id}
        data = payload.model_dump()
        if data["password"] is not None:
            user_orig = await self.get(_id=_id, fields=["backend"])
            if user_orig.backend == "internal":
                data["password"] = self._password(data["password"])
            else:
                data["passwort"] = None

        result = await self._update(query=query, fields=fields, payload=data)
        return ModelV2UserGet(**result)



import logging

from fastapi import APIRouter
from fastapi import Request
from fastapi import HTTPException
from starlette.responses import RedirectResponse

import httpx

from catweazle.crud.users import CrudUsers
from catweazle.crud.oauth import CrudOAuth

from catweazle.errors import ResourceNotFound
from catweazle.errors import AuthenticationError

from catweazle.model.v2.common import ModelV2MetaMulti
from catweazle.model.oauth import ModelOauthProviderGet
from catweazle.model.oauth import ModelOauthProviderGetMulti
from catweazle.model.v2.users import ModelV2UserPut


class ControllerOauthAuthenticate:
    def __init__(
        self,
        log: logging.Logger,
        crud_users: CrudUsers,
        http: httpx.AsyncClient,
        crud_oauth: dict[str, CrudOAuth],
    ):
        self._crud_users = crud_users
        self._http = http
        self._log = log
        self._crud_oauth = crud_oauth
        self._router = APIRouter(
            prefix="/authenticate",
            tags=["authenticate"],
        )

        self.router.add_api_route(
            "/oauth",
            self.get_oauth_providers,
            response_model=ModelOauthProviderGetMulti,
            methods=["GET"],
        )
        self.router.add_api_route(
            "/oauth/{provider}/login",
            self.get_oauth_login,
            methods=["GET"],
        )
        self.router.add_api_route(
            "/oauth/{provider}/auth",
            self.get_oauth_auth,
            methods=["GET"],
        )

    @property
    def crud_oauth(self):
        return self._crud_oauth

    @property
    def crud_users(self):
        return self._crud_users

    @property
    def http(self):
        return self._http

    @property
    def log(self):
        return self._log

    @property
    def router(self):
        return self._router

    async def get_oauth_providers(self):
        providers = list()
        for provider in self.crud_oauth.keys():
            providers.append(ModelOauthProviderGet(id=provider))
        return ModelOauthProviderGetMulti(
            result=providers, meta=ModelV2MetaMulti(result_size=len(providers))
        )

    async def get_oauth_login(self, provider: str, request: Request):
        provider = self.crud_oauth.get(provider, None)
        if provider is None:
            raise HTTPException(status_code=404, detail="oauth provider not found")
        return await provider.oauth_login(request=request)

    async def get_oauth_auth(
        self,
        provider: str,
        request: Request,
    ):
        _provider = self.crud_oauth.get(provider, None)
        if _provider is None:
            raise HTTPException(status_code=404, detail="oauth provider not found")
        token = await _provider.oauth_auth(request=request)
        userinfo = await _provider.get_user_info(token=token["access_token"])
        login = userinfo["login"]
        try:
            user = await self.crud_users.get(_id=login, fields=["backend"])
            if user.backend != f"oauth:{provider}":
                if _provider.backend_override:
                    self.log.warning(
                        f"backend override: backend:{user.backend} -> oauth:{provider}"
                    )
                    await self.crud_users.update(
                        _id=login,
                        payload=ModelV2UserPut(
                            backend=f"oauth{provider}",
                        ),
                        fields=["_id"],
                    )
                else:
                    self.log.error(
                        f"auth backend mismatch: {user.backend} != {provider}"
                    )
                    raise AuthenticationError(
                        msg="backend mismatch, please contact the administrator"
                    )
        except ResourceNotFound:
            await self.crud_users.create_external(
                _id=login,
                payload=ModelV2UserPut(
                    admin=False,
                    email=userinfo["email"],
                    name=userinfo["name"],
                ),
                fields=["_id"],
                backend=f"oauth:{provider}",
            )
        request.session["username"] = login
        return RedirectResponse(url="/")



import logging
import typing

from bson.objectid import ObjectId
from motor.motor_asyncio import AsyncIOMotorCollection
import pymongo
import pymongo.errors

from catweazle.crud.common import CrudMongo

from catweazle.errors import HostNumRangeExceeded

from catweazle.model.v2.common import ModelV2DataDelete
from catweazle.model.v2.common import sort_order_literal
from catweazle.model.v2.instances import ModelV2InstanceGet
from catweazle.model.v2.instances import ModelV2InstanceGetMulti
from catweazle.model.v2.instances import ModelV2instancePost
from catweazle.model.v2.instances import ModelV2instancePut


class CrudInstances(CrudMongo):
    def __init__(
        self, log: logging.Logger, coll: AsyncIOMotorCollection, domain_suffix: str
    ):
        super(CrudInstances, self).__init__(log=log, coll=coll)
        self._domain_suffix = domain_suffix

    @property
    def domain_suffix(self):
        return self._domain_suffix

    async def _next_num(self, indicator):
        instances = await self.search(dns_indicator=indicator)
        taken = list()
        for instance in instances.result:
            taken.append(instance.fqdn)

        fqdn = "{0}{1}".format(indicator, self.domain_suffix)
        host_max_range = 1000
        for number in range(1, host_max_range):
            number = str(number)
            if fqdn.replace("NUM", number) not in taken:
                return number
        raise HostNumRangeExceeded(max_range=host_max_range)

    async def index_create(self) -> None:
        self.log.info(f"creating {self.resource_type} indices")
        await self.coll.create_index([("id", pymongo.ASCENDING)], unique=True)
        await self.coll.create_index([("fqdn", pymongo.ASCENDING)], unique=True)
        self.log.info(f"creating {self.resource_type} indices, done")

    async def create(
        self, _id: str, payload: ModelV2instancePost, fields: list
    ) -> ModelV2InstanceGet:
        data = payload.model_dump()
        data["id"] = _id

        fqdn = f"{payload.dns_indicator}{self.domain_suffix}"
        if "NUM" in payload.dns_indicator:
            number = await self._next_num(payload.dns_indicator)
            fqdn = f"{payload.dns_indicator.replace('NUM', number)}{self.domain_suffix}"
        data["fqdn"] = fqdn
        data["ip_address"] = str(payload.ip_address)
        result = await self._create(fields=fields, payload=data)
        return ModelV2InstanceGet(**result)

    async def delete(
        self,
        _id: str,
    ) -> ModelV2DataDelete:
        query = {"id": _id}
        await self._delete(query=query)
        return ModelV2DataDelete()

    async def get(
        self,
        _id: str,
        fields: list,
    ) -> ModelV2InstanceGet:
        query = {"id": _id}
        result = await self._get(query=query, fields=fields)
        return ModelV2InstanceGet(**result)

    async def resource_exists(
        self,
        _id: str,
    ) -> ObjectId:
        query = {"id": _id}
        return await self._resource_exists(query=query)

    async def search(
        self,
        _id: typing.Optional[str] = None,
        dns_indicator: typing.Optional[str] = None,
        ip_address: typing.Optional[str] = None,
        fqdn: typing.Optional[str] = None,
        fields: typing.Optional[list] = None,
        sort: typing.Optional[str] = None,
        sort_order: typing.Optional[sort_order_literal] = None,
        page: typing.Optional[int] = None,
        limit: typing.Optional[int] = None,
        query: typing.Optional[dict] = None,
    ) -> ModelV2InstanceGetMulti:
        if not query:
            query = {}
        self._filter_re(query, "id", _id)
        self._filter_re(query, "dns_indicator", dns_indicator)
        self._filter_re(query, "ip_address", ip_address)
        self._filter_re(query, "fqdn", fqdn)
        result = await self._search(
            query=query,
            fields=fields,
            sort=sort,
            sort_order=sort_order,
            page=page,
            limit=limit,
        )
        return ModelV2InstanceGetMulti(**result)

    async def update(
        self,
        _id: str,
        payload: ModelV2instancePut,
        fields: list,
    ) -> ModelV2InstanceGet:
        query = {"id": _id}
        data = payload.model_dump()
        result = await self._update(query=query, fields=fields, payload=data)
        return ModelV2InstanceGet(**result)

    async def update_ipa_otp(
        self,
        _id: str,
        ipa_otp: str,
        fields: list,
    ) -> ModelV2InstanceGet:
        query = {"id": _id}
        data = {"ipa_otp": ipa_otp}

        result = await self._update(query=query, fields=fields, payload=data)
        return ModelV2InstanceGet(**result)



import re
from typing import get_args as typing_get_args
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional
from pydantic import BaseModel
from pydantic import StrictStr
from pydantic import field_validator
from pydantic.networks import IPv4Address

from catweazle.model.v2.common import ModelV2MetaMulti
from catweazle.config import Config

config = Config()

filter_literal = Literal[
    "id",
    "dns_indicator",
    "fqdn",
    "ip_address",
    "ipa_otp",
    "meta",
]

filter_list = set(typing_get_args(filter_literal))

sort_literal = Literal[
    "id",
    "dns_indicator",
    "fqdn",
    "ip_address",
]


class ModelV2InstanceGet(BaseModel):
    id: Optional[StrictStr] = None
    dns_indicator: Optional[StrictStr] = None
    fqdn: Optional[StrictStr] = None
    ip_address: Optional[StrictStr] = None
    ipa_otp: Optional[StrictStr] = None
    meta: Optional[Dict[str, str]] = None


class ModelV2InstanceGetMulti(BaseModel):
    result: List[ModelV2InstanceGet]
    meta: ModelV2MetaMulti


class ModelV2instancePost(BaseModel):
    dns_indicator: Optional[StrictStr] = None
    ip_address: IPv4Address
    meta: Optional[Dict[str, str]] = None

    @staticmethod
    @field_validator("dns_indicator", mode="before")
    def validate_dns_indicator(value):
        regex = re.compile(config.app.indicatorregex)
        if not regex.match(value):
            raise ValueError(
                f"dns_indicator does not match regex: {config.app.indicatorregex}"
            )
        return value


class ModelV2instancePut(BaseModel):
    meta: Optional[Dict[str, str]] = None



import logging
import typing

from fastapi import Request

from catweazle.crud.users import CrudUsers
from catweazle.crud.credentials import CrudCredentials
from catweazle.crud.permissions import CrudPermissions

from catweazle.errors import AdminError
from catweazle.errors import CredentialError
from catweazle.errors import PermError
from catweazle.errors import ResourceNotFound
from catweazle.errors import SessionCredentialError

from catweazle.model.v2.users import ModelV2UserGet


class Authorize:
    def __init__(
        self,
        log: logging.Logger,
        crud_permissions: CrudPermissions,
        crud_users: CrudUsers,
        crud_users_credentials: CrudCredentials,
    ):
        self._crud_permission = crud_permissions
        self._crud_users = crud_users
        self._crud_users_credentials = crud_users_credentials
        self._log = log

    @property
    def crud_permission(self) -> CrudPermissions:
        return self._crud_permission

    @property
    def crud_users(self) -> CrudUsers:
        return self._crud_users

    @property
    def crud_users_credentials(self):
        return self._crud_users_credentials

    @property
    def log(self):
        return self._log

    async def get_user(self, request: Request) -> ModelV2UserGet:
        user = self.get_user_from_session(request=request)
        if not user:
            user = await self.get_user_from_credentials(request=request)
        if not user:
            raise SessionCredentialError
        user = await self.crud_users.get(_id=user, fields=["id", "admin"])
        return user

    async def get_user_from_credentials(
        self, request: Request
    ) -> ModelV2UserGet | None:
        try:
            self.log.info("trying to get user from credentials")
            user = await self.crud_users_credentials.check_credential(request=request)
            self.log.debug(f"received user {user} from credentials")
            return user
        except (CredentialError, ResourceNotFound):
            self.log.debug("trying to get user from credentials, failed")
            return None

    def get_user_from_session(self, request: Request) -> typing.Optional[str]:
        self.log.debug("trying to get user from session")
        user = request.session.get("username", None)
        if user is None:
            self.log.debug("trying to get user from session, failed")
            return None
        else:
            self.log.debug(f"received user {user} from session")
            return user

    async def require_admin(self, request, user=None) -> ModelV2UserGet:
        if not user:
            user = await self.get_user(request=request)
        if not user.admin:
            raise AdminError
        return user

    async def require_user(self, request) -> ModelV2UserGet:
        user = await self.get_user(request)
        return user

    async def require_permission(
        self, request, permission, user=None
    ) -> ModelV2UserGet:
        if not user:
            user = await self.get_user(request)
        try:
            return await self.require_admin(request=request, user=user)
        except AdminError:
            pass
        permissions = await self.crud_permission.search(
            users=f"^{user.id}$",
            permissions=f"^{permission}$",
            sort="id",
            sort_order="ascending",
            page=0,
            limit=1,
            fields=["id"],
        )
        if not permissions.result:
            raise PermError(permission=permission)
        return user



import asyncio
import logging

import bonsai.asyncio
import bonsai.errors
import bonsai.pool

from catweazle.errors import AuthenticationError
from catweazle.errors import LdapInvalidDN
from catweazle.errors import LdapResourceNotFound
from catweazle.errors import LdapNoBackend


class CrudLdap:
    def __init__(
        self,
        log: logging.Logger,
        ldap_base_dn: str,
        ldap_bind_dn: str,
        ldap_pool: bonsai.asyncio.AIOConnectionPool,
        ldap_url: str,
        ldap_user_pattern: str,
    ):
        self._log = log
        self._ldap_base_dn = ldap_base_dn
        self._ldap_bind_dn = ldap_bind_dn
        self._ldap_pool = ldap_pool
        self._ldap_url = ldap_url
        self._ldap_user_pattern = ldap_user_pattern

    @property
    def log(self):
        return self._log

    @property
    def ldap_base_dn(self):
        return self._ldap_base_dn

    @property
    def ldap_bind_dn(self):
        return self._ldap_bind_dn

    @property
    def ldap_pool(self):
        if not self._ldap_pool:
            raise LdapNoBackend
        return self._ldap_pool

    @property
    def ldap_url(self):
        return self._ldap_url

    @property
    def ldap_user_pattern(self):
        return self._ldap_user_pattern

    async def _ldap_search(
        self,
        base_dn: str,
        scope: bonsai.LDAPSearchScope,
        query: str,
    ):
        counter = self.ldap_pool.max_connection + 3
        while counter >= 0:
            conn = await self.ldap_pool.get()
            try:
                return await conn.search(base_dn, scope, query)
            except bonsai.pool.EmptyPool:
                self.log.warning("ldap pool empty, waiting 1 second")
                await asyncio.sleep(1)
            except bonsai.errors.ConnectionError:
                conn.close()
                if counter == 0:
                    self.log.error("lost ldap connection, no more retries left")
                else:
                    self.log.error(f"lost ldap connection, {counter} retries left")
                    counter -= 1
            finally:
                await self.ldap_pool.put(conn)

    async def check_user_credentials(self, user: str, password: str):
        if not self.ldap_url:
            raise AuthenticationError
        client = bonsai.LDAPClient(self.ldap_url)
        user_name = self.ldap_user_pattern.format(user)
        client.set_credentials("SIMPLE", user_name, password)
        try:
            async with client.connect(is_async=True) as conn:
                user = await conn.search(
                    self.ldap_base_dn,
                    bonsai.LDAPSearchScope.SUBTREE,
                    f"(userPrincipalName={user_name})",
                )
        except bonsai.errors.AuthenticationError:
            raise AuthenticationError
        return user[0]

    async def get_login(self, user: str):
        user_cn, user_base = user.split(",", maxsplit=1)
        user = await self._ldap_search(
            base_dn=user_base, scope=bonsai.LDAPSearchScope.ONELEVEL, query=user_cn
        )
        return user[0]["sAMAccountName"]

    async def get_logins_from_group(self, group: str):
        try:
            group_cn, group_base = group.split(",", maxsplit=1)
        except ValueError:
            raise LdapInvalidDN
        ldap_group = await self._ldap_search(
            base_dn=group_base, scope=bonsai.LDAPSearchScope.ONELEVEL, query=group_cn
        )
        try:
            ldap_group = ldap_group[0]
        except IndexError:
            raise LdapResourceNotFound
        jobs = []
        for user in ldap_group["member"]:
            jobs.append(asyncio.create_task(self.get_login(user=user)))
        if not jobs:
            self.log.warning(f"ldap group has no members: {group}")
            return []
        done, _ = await asyncio.wait(jobs, return_when=asyncio.ALL_COMPLETED)
        logins = []
        for job in done:
            logins.append(job.result()[0])
        return logins



import logging
from typing import List

import httpx
from fastapi import APIRouter

from catweazle.authorize import Authorize

from catweazle.controller.api.v2.authenticate import ControllerApiV2Authenticate
from catweazle.controller.api.v2.instances import ControllerApiV2Instances
from catweazle.controller.api.v2.permissions import ControllerApiV2Permissions
from catweazle.controller.api.v2.users import ControllerApiV2Users
from catweazle.controller.api.v2.users_credentials import (
    ControllerApiV2UsersCredentials,
)

from catweazle.crud.credentials import CrudCredentials
from catweazle.crud.ldap import CrudLdap
from catweazle.crud.foreman import CrudForeman
from catweazle.crud.instances import CrudInstances
from catweazle.crud.permissions import CrudPermissions
from catweazle.crud.users import CrudUsers


class ControllerApiV2:
    def __init__(
        self,
        log: logging.Logger,
        authorize: Authorize,
        crud_ldap: CrudLdap,
        crud_foreman_backends: List[CrudForeman],
        crud_instances: CrudInstances,
        crud_permissions: CrudPermissions,
        crud_users: CrudUsers,
        crud_users_credentials: CrudCredentials,
        http: httpx.AsyncClient,
    ):
        self._router = APIRouter()
        self._log = log

        self.router.include_router(
            ControllerApiV2Authenticate(
                log=log,
                authorize=authorize,
                crud_users=crud_users,
                http=http,
            ).router,
            responses={404: {"description": "Not found"}},
        )

        self.router.include_router(
            ControllerApiV2Instances(
                log=log,
                authorize=authorize,
                crud_instances=crud_instances,
                crud_foreman_backends=crud_foreman_backends,
            ).router,
            responses={404: {"description": "Not found"}},
        )

        self.router.include_router(
            ControllerApiV2Permissions(
                log=log,
                authorize=authorize,
                crud_permissions=crud_permissions,
                crud_ldap=crud_ldap,
            ).router,
            responses={404: {"description": "Not found"}},
        )

        self.router.include_router(
            ControllerApiV2Users(
                log=log,
                authorize=authorize,
                crud_permissions=crud_permissions,
                crud_users=crud_users,
                crud_users_credentials=crud_users_credentials,
            ).router,
            responses={404: {"description": "Not found"}},
        )

        self.router.include_router(
            ControllerApiV2UsersCredentials(
                log=log,
                authorize=authorize,
                crud_users=crud_users,
                crud_users_credentials=crud_users_credentials,
            ).router,
            responses={404: {"description": "Not found"}},
        )

    @property
    def router(self):
        return self._router



from typing import get_args as typing_get_args
from typing import List
from typing import Literal
from typing import Optional
from pydantic import BaseModel
from pydantic import StrictBool
from pydantic import EmailStr
from pydantic import StrictStr

from catweazle.model.v2.common import ModelV2MetaMulti

filter_literal = Literal[
    "id",
    "admin",
    "backend",
    "email",
    "name",
]

filter_list = set(typing_get_args(filter_literal))

sort_literal = Literal[
    "id",
    "admin",
    "email",
    "name",
]


class ModelV2UserGet(BaseModel):
    admin: Optional[StrictBool] = None
    email: Optional[EmailStr] = None
    name: Optional[StrictStr] = None
    id: Optional[StrictStr] = None
    backend: Optional[StrictStr] = None


class ModelV2UserGetMulti(BaseModel):
    result: List[ModelV2UserGet]
    meta: ModelV2MetaMulti


class ModelV2UserPost(BaseModel):
    admin: StrictBool = False
    email: EmailStr
    name: StrictStr
    password: StrictStr


class ModelV2UserPut(BaseModel):
    admin: Optional[StrictBool] = None
    email: Optional[EmailStr] = None
    name: Optional[StrictStr] = None
    password: Optional[StrictStr] = None
    backend: Optional[StrictStr] = None



import logging

from authlib.integrations.starlette_client import OAuth as authlibOauth

import httpx


class CrudOAuth:
    def __init__(
        self,
        log: logging.Logger,
        http: httpx.AsyncClient,
        backend_override: bool,
        name: str,
        oauth: authlibOauth,
        scope: str,
        client_id: str,
        client_secret: str,
        authorize_url: str,
        access_token_url: str,
    ):
        self._backend_override = backend_override
        self._http = http
        self._log = log
        self._name = name
        self._oauth = oauth
        self._scope = scope
        self.oauth.register(
            name=name,
            client_id=client_id,
            client_secret=client_secret,
            authorize_url=authorize_url,
            access_token_url=access_token_url,
            client_kwargs={"scope": scope},
        )

    @property
    def backend_override(self):
        return self._backend_override

    @property
    def http(self):
        return self._http

    @property
    def log(self):
        return self._log

    @property
    def name(self):
        return self._name

    @property
    def oauth(self):
        return self._oauth

    @property
    def scope(self):
        return self._scope

    async def oauth_login(self, request):
        redirect_url = request.url_for("get_oauth_auth", provider=self.name)
        provider = self.oauth.create_client(name=self.name)
        return await provider.authorize_redirect(request, str(redirect_url))

    async def oauth_auth(self, request):
        provider = self.oauth.create_client(name=self.name)
        token = await provider.authorize_access_token(request)
        return token

    async def get_user_info(self, token):
        raise NotImplementedError


class CrudOAuthGitHub(CrudOAuth):
    def __init__(
        self,
        log: logging.Logger,
        http: httpx.AsyncClient,
        backend_override: bool,
        name: str,
        oauth: authlibOauth,
        scope: str,
        client_id: str,
        client_secret: str,
        authorize_url: str,
        access_token_url: str,
        userinfo_url: str,
    ):
        super(CrudOAuthGitHub, self).__init__(
            log=log,
            http=http,
            backend_override=backend_override,
            name=name,
            oauth=oauth,
            scope=scope,
            client_id=client_id,
            client_secret=client_secret,
            authorize_url=authorize_url,
            access_token_url=access_token_url,
        )

        self._userinfo_url = userinfo_url

    @property
    def userinfo_url(self):
        return self._userinfo_url

    async def get_user_info(self, token: str):
        user_info = await self.http.get(
            url=self.userinfo_url, headers={"Authorization": f"token {token}"}
        )
        return user_info.json()



import logging
from typing import Set

from fastapi import APIRouter
from fastapi import Query
from fastapi import Request

from catweazle.authorize import Authorize

from catweazle.crud.credentials import CrudCredentials
from catweazle.crud.users import CrudUsers

from catweazle.model.v2.common import ModelV2DataDelete
from catweazle.model.v2.common import sort_order_literal
from catweazle.model.v2.credentials import filter_list
from catweazle.model.v2.credentials import filter_literal
from catweazle.model.v2.credentials import sort_literal
from catweazle.model.v2.credentials import ModelV2CredentialGet
from catweazle.model.v2.credentials import ModelV2CredentialGetMulti
from catweazle.model.v2.credentials import ModelV2CredentialPost
from catweazle.model.v2.credentials import ModelV2CredentialPostResult
from catweazle.model.v2.credentials import ModelV2CredentialPut


class ControllerApiV2UsersCredentials:
    def __init__(
        self,
        log: logging.Logger,
        authorize: Authorize,
        crud_users: CrudUsers,
        crud_users_credentials: CrudCredentials,
    ):
        self._authorize = authorize
        self._crud_users = crud_users
        self._crud_users_credentials = crud_users_credentials
        self._log = log
        self._router = APIRouter(
            prefix="/users/{user_id}/credentials",
            tags=["users_credentials"],
        )

        self.router.add_api_route(
            "",
            self.create,
            response_model=ModelV2CredentialPostResult,
            response_model_exclude_unset=True,
            methods=["POST"],
            status_code=201,
        )
        self.router.add_api_route(
            "",
            self.search,
            response_model=ModelV2CredentialGetMulti,
            response_model_exclude_unset=True,
            methods=["GET"],
        )
        self.router.add_api_route(
            "/{credential_id}",
            self.delete,
            response_model=ModelV2DataDelete,
            response_model_exclude_unset=True,
            methods=["DELETE"],
        )
        self.router.add_api_route(
            "/{credential_id}",
            self.get,
            response_model=ModelV2CredentialGet,
            response_model_exclude_unset=True,
            methods=["GET"],
        )
        self.router.add_api_route(
            "/{credential_id}",
            self.update,
            response_model=ModelV2CredentialGet,
            response_model_exclude_unset=True,
            methods=["PUT"],
        )

    @property
    def authorize(self):
        return self._authorize

    @property
    def crud_users(self):
        return self._crud_users

    @property
    def crud_users_credentials(self):
        return self._crud_users_credentials

    @property
    def log(self):
        return self._log

    @property
    def router(self):
        return self._router

    async def create(
        self,
        data: ModelV2CredentialPost,
        user_id: str,
        request: Request,
    ):
        if user_id == "_self":
            user = await self.authorize.get_user(request=request)
            user_id = user.id
        else:
            await self.authorize.require_admin(request=request)
        await self.crud_users.resource_exists(_id=user_id)
        return await self.crud_users_credentials.create(
            owner=user_id,
            payload=data,
        )

    async def delete(
        self,
        user_id: str,
        credential_id: str,
        request: Request,
    ):
        if user_id == "_self":
            user = await self.authorize.get_user(request=request)
            user_id = user.id
        else:
            await self.authorize.require_admin(request=request)
        return await self.crud_users_credentials.delete(
            _id=credential_id, owner=user_id
        )

    async def get(
        self,
        request: Request,
        user_id: str,
        credential_id: str,
        fields: Set[filter_literal] = Query(default=filter_list),
    ):
        if user_id == "_self":
            user = await self.authorize.get_user(request=request)
            user_id = user.id
        else:
            await self.authorize.require_admin(request=request)
        return await self.crud_users_credentials.get(
            owner=user_id, _id=credential_id, fields=list(fields)
        )

    async def search(
        self,
        request: Request,
        user_id: str,
        fields: Set[filter_literal] = Query(default=filter_list),
        sort: sort_literal = Query(default="id"),
        sort_order: sort_order_literal = Query(default="ascending"),
        page: int = Query(default=0, ge=0, description="pagination index"),
        limit: int = Query(
            default=10,
            ge=10,
            le=1000,
            description="pagination limit, min value 10, max value 1000",
        ),
    ):
        if user_id == "_self":
            user = await self.authorize.get_user(request=request)
            user_id = user.id
        else:
            await self.authorize.require_admin(request=request)
        result = await self._crud_users_credentials.search(
            owner=user_id,
            fields=list(fields),
            sort=sort,
            sort_order=sort_order,
            page=page,
            limit=limit,
        )
        return result

    async def update(
        self,
        request: Request,
        user_id: str,
        credential_id: str,
        data: ModelV2CredentialPut,
        fields: Set[filter_literal] = Query(default=filter_list),
    ):
        if user_id == "_self":
            user = await self.authorize.get_user(request=request)
            user_id = user.id
        else:
            await self.authorize.require_admin(request=request)
        return await self.crud_users_credentials.update(
            _id=credential_id, owner=user_id, payload=data, fields=list(fields)
        )
